#ifndef __NTR_INLINETANK_H__
#define __NTR_INLINETANK_H__

#define STRICT

#include "orbitersdk.h"
//#include "autopilot.h"
#include <math.h>
#include <vector>
//#include "rotations.h"
//#include "PIDControl.h"
//#include "nlopt.hpp"

// Screen output
#define GREEN		RGB(0x00, 0xFF, 0x00)
#define YELLOW		RGB(0xDF, 0xDF, 0x00)
#define	WHITE		RGB(0xFF, 0xFF, 0xFF)
#define RED			RGB(0xFF, 0x00, 0x00)
#define GREY		RGB(0xE0, 0xE0, 0xE0)
#define GOLD		RGB(0xFF, 0xFF, 0x00)

static int lineY[] = {152, 168, 184, 200, 216, 232, 248, 264, 280, 296, 312, 328, 344, 360, 376, 392, 408};
#define HUDTextOut_CLR(skp, x, y, colour, text, ...)	{\
	DWORD old_clr = skp->SetTextColor(colour); \
	char buf[1024]; \
	sprintf_s(buf, 1024, text, __VA_ARGS__); \
	skp->Text (x+roxl, y+royl,buf, (int)strlen(buf));\
	skp->SetTextColor(old_clr);\
}

enum glass_screen_t
{
	LAUNCH = 0,
	CHECKOUT,
	ENGINE_CONTROL,
	DAP_CONFIG,
	DOCKING,
	COMMUNICATIONS
};

//NTR_inlinetank.msh
#define MESH_MAIN (0)

// mesh groups - update after each edit!
#define MESHGRP_TANK1 (0)
// docking lime
#define MESHGRP_TANK2 (1)
// docking red
#define MESHGRP_TANK3 (2)
#define MESHGRP_SARJ1 (4)
#define MESHGRP_SARJ2 (3)
#define MESHGRP_ARRAY2 (5)
#define MESHGRP_ARRAY1 (6)
#define MESHGRP_ARRAY2_2 (10)
#define MESHGRP_ARRAY1_2 (9)
#define MESHGRP_TANK4 (7)
#define MESHGRP_CUBE13 (8)

#define MAT_ActiveMLI2  (0)
#define MAT_DockingLime (1)
#define MAT_DockingRed  (2)
#define MAT_SARJ_black  (3)
#define MAT_SolarMat1   (4)
#define MAT_default2     (5)
#define MAT_default3     (6)
#define MAT_default4     (7)
#define MAT_default5     (8)


//DESIGN PARAMS
// ==============================================================
// Vessel parameters
// ==============================================================
const double  PB_SIZE       = 6.7;             // mean radius [m] from volume from shipedit
const double  NTR_CLIPRADIUS = 16.0;            // to avoid artefacts

//FIXME!!
const VECTOR3 PB_CS         = {147.3,147.3, 85.39}; // x,y,z cross sections [m^2] from shipedit
const VECTOR3 PB_PMI_FULL   = {27.25, 27.25, 9.96};// principal moments of inertia (mass-normalised) [m^2] - from shipedit
const VECTOR3 PB_PMI_EMPTY  = {16.91, 16.91, 8.25};// principal moments of inertia (mass-normalised) [m^2] - from shipedit, empty!
const VECTOR3 PB_RD         = {0.025,0.025,0.02};//{0.05,0.1,0.05};  // rotation drag coefficients :) FIXME

// MASS
const double  PB_EMPTYMASS  = 21400.0;           // empty vessel mass [kg]
const double  PB_LH2MASS    = 69900.0;           // max fuel mass [kg]
const double  PB_RCSFUELMASS =  100.0;           // boiloff gas - for now! //91.4- etc.

const VECTOR3 PB_TDP[3]     = {{0,-1.5, 0 },{-1,-1.5,0},{1,-1.5,0}}; // touchdown points [m] - needed for COG management!
const VECTOR3 PB_COP        = {0,0,0};//{0,0,-0.1};      // centre of pressure for airfoils [m] :)
const double  PB_VLIFT_C    = 2.0;             // chord length [m]
const double  PB_VLIFT_S    = 2.0;             // wing area [m^2]
const double  PB_VLIFT_A    = 2.5;             // wing aspect ratio
const double  PB_HLIFT_C    = 2.0;             // chord length [m]
const double  PB_HLIFT_S    = 1.5;             // wing area [m^2]
const double  PB_HLIFT_A    = 2.0;             // wing aspect ratio

const DWORD NTR_Transponder = 518;
const DWORD NTR_IDS_Channel1 = 522;
const DWORD NTR_IDS_Channel2 = 523;

const VECTOR3 PB_DOCK1_POS   = {0,0,8.15};      // docking port location [m]
const VECTOR3 PB_DOCK1_DIR   = {0,0,1};         // docking port approach direction
const VECTOR3 PB_DOCK1_ROT   = {0,1,0};        // docking port alignment direction

const VECTOR3 PB_DOCK2_POS   = {0,0,-8.15};      // docking port location [m]
const VECTOR3 PB_DOCK2_DIR   = {0,0,-1};         // docking port approach direction
const VECTOR3 PB_DOCK2_ROT   = {0,-1,0};        // docking port alignment direction

// RCS jets
const double  PB_ISP        = 150*G;             // RCS fuel-specific impulse [m/s] - cold (boiloff) gas from the tank
const double  PB_MAXRCSTH   = 50; //[N]

const int NUM_RCS_THRUSTERS = 18;
const int MAX_NUM_THRUSTERS_PER_GROUP = 4;

#define RCS_XOFFSET (4.5)
#define RCS_YOFFSET (4.5)
#define RCS_ZOFFSET (4.5)

const VECTOR3 V_RCS_POS[] = {
	_V( RCS_XOFFSET,0, RCS_ZOFFSET), 
	_V( RCS_XOFFSET,0, RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0, RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0, RCS_ZOFFSET),
	_V( RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V( RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V( RCS_XOFFSET,0, RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0, RCS_ZOFFSET),
	_V( RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0,-RCS_ZOFFSET),
	// +Z
	_V(-RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(+RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(0,-RCS_YOFFSET,-RCS_ZOFFSET),
	_V(0,+RCS_YOFFSET,-RCS_ZOFFSET),
	// -Z
	_V(-RCS_XOFFSET,0,+RCS_ZOFFSET),
	_V(+RCS_XOFFSET,0,+RCS_ZOFFSET)};

const VECTOR3 V_RCS_DIR[] = {
_V(0, 1,0), //+Y
_V(0,-1,0), //-Y
_V(0, 1,0), //+Y
_V(0,-1,0), //-Y
_V(0, 1,0), //+Y
_V(0,-1,0), //-Y
_V(0, 1,0), //+Y
_V(0,-1,0), //-Y
_V(-1,0,0), //-X
_V( 1,0,0), //+X
_V(-1,0,0), //-X
_V( 1,0,0), //+X
_V(0,0, 1), //+Z
_V(0,0, 1), //+Z
_V(0,0, 1), //+Z
_V(0,0, 1), //+Z
_V(0,0,-1), //-Z
_V(0,0,-1)};//-Z


const double RCS_DAP_Gain = 5.0; //not so free... set to 3 and it's fine?

//
const double DB_FINE = 0.05;
const double DB_MEDIUM = 0.5; //deg
const double DB_COARSE = 5.0; //deg

const double RATE_SLOWEST = 0.05; //deg/second
const double RATE_SLOW = 0.2;
const double RATE_MEDIUM = 0.5;
const double RATE_FAST = 2.0;

const double DB_OFFSET = 0.044; //degrees
const double DB_SLOPE = -0.6; //deg/sec/degree
const double DB_HSLOPE = -0.8; //deg/sec/degree
const double DB_WL = 0.4; //deg/sec -from Apollo DAP
const double DB_RATE = 0.2; //deg/sec


const int NUM_CONTROLLERS = 3;

#define RCS_P (0)
#define RCS_Y (1)
#define RCS_R (2)

//Self-destruct and transition criteria
const double LAUNCH_G_LIMIT = 7.0;
const int    LAUNCH_STATIC_dP_dt_STEP_LIMIT = 3;
const int    LAUNCH_G_STEP_LIMIT = 5;
const double LAUNCH_STATIC_dP_dt_THRESHOLD = 100.0;

const double MAX_DYN_PRESSURE = 1000; //1 kPa - don't aerobrake!

const double SAFE_PeA = 280.0;
const double SAFE_ApA = 420.0;

//Animations
const int NUM_ANIMATIONS = 2;
#define SOLAR1_ANIMATION (0)
#define SOLAR2_ANIMATION (1)

const VECTOR3 PVA1_Axis_Offset = _V( 3.189052, 3.189052,13.45);
const VECTOR3 PVA2_Axis_Offset = _V(-3.189052,-3.189052,13.45);
const VECTOR3 DIR_XY_225_Z_0 = _V(-0.707107,+0.707107,0);
const VECTOR3 DIR_XY_045_Z_0 = _V(+0.707107,-0.707107,0);

// animation transformation definitions
static UINT GRP_LWING = 2;
static UINT GRP_RWING = 3;


static VECTOR3 LWING_REF  = {-1.3,-0.725,-1.5};
static VECTOR3 LWING_AXIS = {-0.9619,-0.2735,0};
static VECTOR3 RWING_REF  = {1.3,-0.725,-1.5};
static VECTOR3 RWING_AXIS = {0.9619,-0.2735,0};
static float AILERON_RANGE = (float)(20.0*RAD);
static float ELEVATOR_RANGE = (float)(30.0*RAD);

const double  RCS_EXHAUST_LENGTH = 1.0;

// Grail-shaped beacon...
static VECTOR3 beaconPositions[] = {_V(3.0, 0.0, 8.15),
									_V(-3.0, 0.0, 8.15),
									_V(0.0, 3.0, 8.15),
									_V(0.0, -3.0, 8.15),
									_V(3.0, 0.0, -8.15),
									_V(-3.0, 0.0, -8.15),
									_V(0.0, 3.0, -8.15),
									_V(0.0, -3.0, -8.15) };


//static VECTOR3 beaconPlusXpos = _
//static VECTOR3 beaconMinusXpos = 
//static VECTOR3 beaconPlusYpos = _
//static VECTOR3 beaconMinusYpos = 
//static VECTOR3 beacon2PlusXpos = 
//static VECTOR3 beacon2MinusXpos =
//static VECTOR3 beacon2PlusYpos = 
//static VECTOR3 beacon2MinusYpos =
static VECTOR3 beaconColourRed = _V(1.0, 0.0, 0.0);
static VECTOR3 beaconColourGreen = _V(0.0, 1.0, 0.0);
static VECTOR3 beaconColourBlue = _V(0.0, 0.0, 1.0);
static VECTOR3 beaconColourWhite = _V(1.0, 1.0, 1.0);

const int NUM_BEACONS = 8;
#define BEACON_PLUSX  (0)
#define BEACON_MINUSX (1)
#define BEACON_PLUSY  (2)
#define BEACON_MINUSY (3)
#define BEACON2_PLUSX  (4)
#define BEACON2_MINUSX (5)
#define BEACON2_PLUSY  (6)
#define BEACON2_MINUSY (7)

enum simple_anim_status_t
{
	OPEN = 0,
	CLOSED,
	OPENING,
	CLOSING,
	LAST
};


//CLASS INTERFACE
// ==============================================================
// NTR stage class interface
// ==============================================================

class NTR_inlinetank: public VESSEL3 {
public:
	NTR_inlinetank (OBJHANDLE hVessel, int flightmodel);
	~NTR_inlinetank ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void clbkPreStep (double simt, double simdt, double mjd);
	void clbkPostStep (double simt, double simdt, double mjd);

	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	void clbkSaveState  (FILEHANDLE  scn);
	int clbkConsumeBufferedKey  ( DWORD  key,  bool  down,  char *  kstate );
	void clbkDockEvent  ( int  dock,  OBJHANDLE  mate);
	void clbkPostCreation  (); 
	int clbkGeneric(int msgid, int prm, void *context);
	void clbkVisualCreated  ( VISHANDLE  vis,  int  refcount);
	void clbkVisualDestroyed  ( VISHANDLE  vis, int  refcount);

	double GetStackMass();
	bool GetStackGroundContact();

//	void SetRCSstringLevel_SingleStep(const stack_RCS_string_t& s, const double dlevel) const;
	void SetThrusterGroupLevel_SingleStep(THGROUP_TYPE  thgt,  double  dlevel) const;

	void ComputeOptimumRCS(double refThrust);

	void SelfDestruct(const char * reason);

	void DefineAnimations();

//	double thermalPower, currentThrust, currentIsp, reactorPeriod_s; 
	//bool Startup, Shutdown, EngineOn, NominalPower, Meltdown;
	// valves? probabilities of failures?
//	reactor_state_t StageState;

//	bool MaintainCommandedRates;

//	double dV_sinceBurnStart, massAtBurnStart;
//	double XePoisoning;
	//INDIVIDUAL
//	double engXePoisoning[NUM_REACTORS];

//	vector<stackVessel_t> stackInfo;
//	vector<drawdown_t> drawdownSequence;
//	THRUSTER_HANDLE th_NTR[3];
	PROPELLANT_HANDLE hpr_LH2, hpr_RCS;
//	double commanded_dV;
//	double NTRGimbal[NUM_REACTORS][2];
//	bool GimbalSelect[NUM_REACTORS];
//	double EngineThrust[NUM_REACTORS];
	//INDIVIDUAL
//	double engCurrentIsp[NUM_REACTORS];
//	double engReactorPeriod_s[NUM_REACTORS]; 
//	double ReactorPower[NUM_REACTORS];
//	reactor_state_t ReactorState[NUM_REACTORS];
//	turbopump_state_t TPAState[NUM_REACTORS][2];
//	PIDControl Controllers[NUM_CONTROLLERS]; //TVCControlP,TVCControlY, RCSControlR, RCSControlP, RCSControlY;
//	bool FuelLevelTripActivated;
	bool IsInLaunchMode;
	bool IsInCheckoutMode;
	int Static_dP_dt_criterion_counter;
	int G_limit_criterion_counter;
	double old_STP;
	double NTR_Life_s; // seconds
	simple_anim_status_t AnimationStatus[NUM_ANIMATIONS];
	static double AnimationSpeeds[NUM_ANIMATIONS];
	double AnimationDeployments[NUM_ANIMATIONS];
	static UINT Animations[NUM_ANIMATIONS];
	VECTOR3 CommandedRates;
	VECTOR3 DAP_Deadbands; //TODO!
	// DONE: gains for the TVC and RCS controllers TODO: saving/reloading them
	bool ActiveDAPChannels[3];

	bool ShallTrackSun;
	
	bool ShallTrackPoint;
	VECTOR3 TrackPointEulers;
//	double UllageBurnTimeLeft;
	glass_screen_t CurrentScreen;
//	stack_RCS_config_t stack_RCS;
//	void OptimizeTorque(const vector<double> & thrust, vector<double> &pitch, vector<double> &yaw); //
//	double RateFromAttError(const double AttError, const double Deadband);
private:
	static void vlift (VESSEL *v, double aoa, double M, double Re,
		void *context, double *cl, double *cm, double *cd);
	static void hlift (VESSEL *v, double aoa, double M, double Re,
		void *context, double *cl, double *cm, double *cd);

	// transformations for control surface animations - there are no FCS here, but saved for safety anyway
	//static MGROUP_ROTATE trans_Laileron, trans_Raileron;
	//static MGROUP_ROTATE trans_Lelevator, trans_Relevator;

	BEACONLIGHTSPEC beacons[NUM_BEACONS];

//	void FillDrawdownSequence_and_OptimizeRCS();
//	void FillDrawdownSequence();

	VISHANDLE visual;
	UINT MainMeshID;
//	double fullCOG_elev;
//	double emptyCOG_elev;
//	bool shallFillDrawdownSeq;
//	nlopt::opt* optProblem;
};


#endif //!__NTR_INLINETANK_H__