#ifndef __NTR_TEST_H__
#define __NTR_TEST_H__

#define STRICT

#include "orbitersdk.h"
#include "autopilot.h"
#include <math.h>
#include <vector>
#include "rotations.h"
#include "PIDControl.h"
#include "nlopt.hpp"
#include "solarsys.h"
#include "eclipse.h"
#include "thermal.h"
#include "brent2.h"

#include "NTR_test_mesh.h"

// Screen output
#define GREEN		RGB(0x00, 0xFF, 0x00)
#define YELLOW		RGB(0xDF, 0xDF, 0x00)
#define	WHITE		RGB(0xFF, 0xFF, 0xFF)
#define RED			RGB(0xFF, 0x00, 0x00)
#define GREY		RGB(0xE0, 0xE0, 0xE0)
#define GOLD		RGB(0xFF, 0xFF, 0x00)

static int lineY[] = {152, 168, 184, 200, 216, 232, 248, 
					  264, 280, 296, 312, 328, 344, 360, 
					  376, 392, 408, 424, 440, 456, 472, 
					  488, 504, 520, 536, 552, 568, 584};

#define HUDTextOut_CLR(skp, x, y, colour, text, ...)	{\
	DWORD old_clr = skp->SetTextColor(colour); \
	char buf[1024]; \
	sprintf_s(buf, 1024, text, __VA_ARGS__); \
	skp->Text (x+roxl, y+royl,buf, (int)strlen(buf));\
	skp->SetTextColor(old_clr);\
}

//NTR_test.msh
#define MESH_MAIN (0)
//// mesh groups - update after each edit!
//
//#define MESHGRP_CommonTank_AftRadiator  (0)
//#define MESHGRP_CommonTank_DockingLime  (1)
//#define MESHGRP_CommonTank_DockingRed   (2)
//#define MESHGRP_NTR_Charlo_EngineMat    (3)
//#define MESHGRP_NTR_Betty_EngineMat     (4)
//#define MESHGRP_NTR_Alice_EngineMat     (5)
//#define MESHGRP_CommonTank_NTR_MLI      (6)
//#define MESHGRP_NTR_Charlo_NozzleInside (7)
//#define MESHGRP_NTR_Betty_NozzleInside  (8)
//#define MESHGRP_NTR_Alice_NozzleInside  (9)
//#define MESHGRP_SUNSENS_AFT_2 (10)
//#define MESHGRP_SUNSENS_AFT_1 (11)
//#define MESHGRP_SUNSENS_FWD_2 (12)
//#define MESHGRP_SARJ1 (13)
//#define MESHGRP_SARJ2 (14)
//#define MESHGRP_SUNSENS_FWD_1 (15)
//#define MESHGRP_ARRAY2 (16)
//#define MESHGRP_ARRAY1 (17)
//#define MESHGRP_TANK_TOP (18)
//#define MESHGRP_ARRAY2_2 (19)
//#define MESHGRP_ARRAY1_2 (20)
//#define MESHGRP_TANK4 (21)
//#define MESHGRP_SBAND1  (22)
//#define MESHGRP_CUBE13 (23)
//#define MESHGRP_RCS_T4 (24)
//#define MESHGRP_RCS_T3 (25)
//#define MESHGRP_RCS_T2 (26)
//#define MESHGRP_RCS_T1 (27)
//#define MESHGRP_RCS_B4 (28)
//#define MESHGRP_RCS_B3 (29)
//#define MESHGRP_RCS_B2 (30)
//#define MESHGRP_RCS_B1 (31)
//
//#define MAT_AftRadiator    (0)
//#define MAT_DockingLime  (1)
//#define MAT_DockingRed   (2)
//#define MAT_EngineMat    (3)
//#define MAT_NTR_MLI      (4)
//#define MAT_NozzleInside (5)
//#define MAT_SARJ_black   (6)
//#define MAT_SolarTexture (7)
//#define MAT_TankTop      (8)
//#define MAT_default      (9)


//DESIGN PARAMS
// ==============================================================
// Vessel parameters
// ==============================================================
const double  PB_SIZE       = 6.7;             // mean radius [m] from volume from shipedit
const double  NTR_CLIPRADIUS = 16.0;            // to avoid artefacts

const double NTR_cyl_length   = 16.63; //meters
const double NTR_cyl_diameter = 9.02; // meters
const double NTR_radiator_area = 0.85*PI*NTR_cyl_diameter*(NTR_cyl_length+NTR_cyl_diameter/2.0);//15% for engines and stuff

const VECTOR3 PB_CS         = {199.2,199.2,62.39}; // x,y,z cross sections [m^2] from shipedit
const VECTOR3 PB_PMI_FULL   = {53.58, 53.58, 9.31};// principal moments of inertia (mass-normalised) [m^2] - from shipedit
const VECTOR3 PB_PMI_EMPTY  = {16.91, 16.91, 8.25};// principal moments of inertia (mass-normalised) [m^2] - from shipedit, empty!
const VECTOR3 PB_RD         = {0.025,0.025,0.02};//{0.05,0.1,0.05};  // rotation drag coefficients :) FIXME

// MASS
const double  PB_EMPTYMASS  = 33700.0;           // empty vessel mass [kg]
const double  PB_LH2MASS    = 59400.0;           // max fuel mass [kg]
const double  PB_RCSFUELMASS = 3600.0;           // hypergolic fuel for RCS thrusters

const VECTOR3 PB_TDP[3]     = {{0,-1.5, 0 },{-1,-1.5,0},{1,-1.5,0}}; // touchdown points [m] - needed for COG management!
const VECTOR3 PB_COP        = {0,0,0};//{0,0,-0.1};      // centre of pressure for airfoils [m] :)
const double  PB_VLIFT_C    = 2.0;             // chord length [m]
const double  PB_VLIFT_S    = 2.0;             // wing area [m^2]
const double  PB_VLIFT_A    = 2.5;             // wing aspect ratio
const double  PB_HLIFT_C    = 2.0;             // chord length [m]
const double  PB_HLIFT_S    = 1.5;             // wing area [m^2]
const double  PB_HLIFT_A    = 2.0;             // wing aspect ratio

const DWORD NTR_Transponder = 515;
const DWORD NTR_IDS_Channel = 520;


const double  PB_MAXMAINTH  = 3e4;             // NOTIONAL DUMMY ENGINE

const VECTOR3 PB_DOCK_POS   = {0,0,14.05};      // docking port location [m]
const VECTOR3 PB_DOCK_DIR   = {0,0,1};         // docking port approach direction
const VECTOR3 PB_DOCK_ROT   = {0,1,0};        // docking port alignment direction

// MAIN ENGINES
const int NUM_REACTORS = 3;

#define NTR_Alice (0)
#define NTR_Betty (1)
#define NTR_Clara (2)

const VECTOR3 V_Alice = _V(       0.0, -1.484836,     -15.35);
const VECTOR3 V_Betty = _V(  1.283126,  0.742418,     -15.35);
const VECTOR3 V_Clara = _V( -1.283126,  0.742418,     -15.35);

const double  PB_NTR_ISP      = 819*G;             // NTR - specific impulse [m/s] - should be 830?
const double  NTR_MAXTHRUST   = 111000; // [N] - 25 klbf, not 15
const double  ZeroThrustLevel = 1000.0; //[N] - where have I got that!?

const double NTR_MAX_LIFE = 5.0*3600.0; // 5 hours of max temp work (proportional extension at lower temps)

// RCS jets
const double  PB_ISP        = 320*G;             // RCS fuel-specific impulse [m/s] - this number ain't yet ready for flight? 
const double  PB_MAXRCSTH   = 200; //[N] however standard RCS thrusters are around 450 N (Wertz 1999)

const int NUM_RCS_THRUSTERS = 18;
const int MAX_NUM_THRUSTERS_PER_GROUP = 4;

#define RCS_XOFFSET (4.5)
#define RCS_YOFFSET (4.5)
#define RCS_ZOFFSET (4.5)

const VECTOR3 V_RCS_POS[] = {
	_V( RCS_XOFFSET,0, RCS_ZOFFSET), 
	_V( RCS_XOFFSET,0, RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0, RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0, RCS_ZOFFSET),
	_V( RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V( RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V( RCS_XOFFSET,0, RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0, RCS_ZOFFSET),
	_V( RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(-RCS_XOFFSET,0,-RCS_ZOFFSET),
	// +Z
	_V(-RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(+RCS_XOFFSET,0,-RCS_ZOFFSET),
	_V(0,-RCS_YOFFSET,-RCS_ZOFFSET),
	_V(0,+RCS_YOFFSET,-RCS_ZOFFSET),
	// -Z
	_V(-RCS_XOFFSET,0,+RCS_ZOFFSET),
	_V(+RCS_XOFFSET,0,+RCS_ZOFFSET)};

const VECTOR3 V_RCS_DIR[] = {
_V(0, 1,0), //+Y
_V(0,-1,0), //-Y
_V(0, 1,0), //+Y
_V(0,-1,0), //-Y
_V(0, 1,0), //+Y
_V(0,-1,0), //-Y
_V(0, 1,0), //+Y
_V(0,-1,0), //-Y
_V(-1,0,0), //-X
_V( 1,0,0), //+X
_V(-1,0,0), //-X
_V( 1,0,0), //+X
_V(0,0, 1), //+Z
_V(0,0, 1), //+Z
_V(0,0, 1), //+Z
_V(0,0, 1), //+Z
_V(0,0,-1), //-Z
_V(0,0,-1)};//-Z

// the worst case is pogo effect at burn...

const double NTR_MAX_INTERTANK_FLOW = 48.0; //kg/s of LH2 (?) 
const double UllageBurnPeriod_s = 20.0; // TODO: for now, make it variable later!

//POWER!
const double SOLAR_BACK_FACE_PENALTY = 0.3;
const double SOLAR_ARRAY_AREA = 22.6; // rough
const double SOLAR_EFFICIENCY = 0.22; //Multijunction at BOL. GaInP/GaAs
const double SOLAR_DEGRADATION_PER_YEAR = 0.005; // per cent
// how much time will be spent in Van Allen belts?

// how do we size the batteries?

// parallel charging : use LC3 chargers?

const double DISTRIBUTION_EFFICIENCY = 0.95; //??
//const double SOLAR_DISTRIBUTION_EFFICIENCY = 0.91; //??
//const double BATTERY_DISTRIBUTION_EFFICIENCY = 0.93; //??
const double CHARGER_EFFICIENCY = 0.85; //??
// what current? @13 amp, 27 Vdc = 351
const double RECHARGE_RATE = 351*4; //?? 70% of the recharge period is fast charging at 1C, 30% is Imin, constant voltage TBI
// HEHE: the recharge is way too slow (aren't we charging several cells at once?

const double LiIonDepthOfDischarge = 0.7;//DRA 5.0 100 W-h/kg
const double BATTERY_INITIAL_CAPACITY = 3200; //will fine-tune later. 
//average eclipse load times eclipse duration times battery-to-load efficiency over DOD
//say, 3 kW * 38 minutes (@500 km) / (0.93*0.7)

//Cryomech Inc. single stage GM cryofridges: AL230: 30W at 20K, input power 5.4kW @ 60Hz - 1:180! (at 17 is 275 though)

const double CryocoolerSpecificPower = 275.0; //W/W
const double CryocoolerMinimumPower = 2600.0;//??
const double CryocoolerMaximumPower = 7000.0;//??7 kW ??? to be dissipated...
// the DRA calls for 15 kWe cryocoolers in the assembled stack, I made it 7kW here

const double CDH_Power = 7.0; //Watt. with telemetry from the reactors and all subsystems etc.

// also possible to dump solar arrays on the core stage just before TEI.

const int Num_MLI_Layers = 220; // thick-skinned bastard.
const double MLI_layer_thickness = 0.025/110.0;//m
const double MLI_eq_eff_conductivity = 0.00001;//W/m-K 

// It could be worth it to consider two spheres... - NOT.

const double LH2_tank_diameter = 8.8;
const double LH2_tank_radius = 0.5*LH2_tank_diameter;
const double LH2_tank_cyl_Len = 13.6;
const double LH2_tank_cyl_V = PI*(LH2_tank_radius*LH2_tank_radius)*LH2_tank_cyl_Len;
const double LH2_tank_cyl_area=2.0*PI*LH2_tank_radius*LH2_tank_cyl_Len;
const double LH2_tank_endSeg_radius = LH2_tank_radius*sqrt(2.0);
const double LH2_tank_endSeg_height = LH2_tank_endSeg_radius-sqrt(LH2_tank_endSeg_radius*LH2_tank_endSeg_radius-(LH2_tank_radius*LH2_tank_radius));
const double LH2_tank_endSeg_V = PI*LH2_tank_endSeg_height*LH2_tank_endSeg_height*(3.0*LH2_tank_endSeg_radius-LH2_tank_endSeg_height)/3.0;
const double LH2_tank_endSeg_area = PI*(LH2_tank_endSeg_height*LH2_tank_endSeg_height+2.0*LH2_tank_radius*LH2_tank_radius);
const double LH2_tank_V = LH2_tank_cyl_V+2.0*LH2_tank_endSeg_V;
const double LH2_tank_area = LH2_tank_cyl_area+2.0*LH2_tank_endSeg_area;


const double BATTERY_ACTIVATION_THRESHOLD = 2600.0;

// to realistically get heat loads in LEO and in the orbit around Mars, the following things should be considered:
// 1. direct solar loads (ADD ECLIPSE!)

// solar backface heats the tank if Earthlight reflects off it. - TODO!

// 8-mil quartz mirrors all around! if you choose anything else, you'll be sorry
const double EXTERNAL_SURFACE_SOLAR_ABSORPTANCE = 0.08;
const double EXTERNAL_SURFACE_IR_EMISSIVITY = 0.80;

enum NTR_Component_t {
	CDH_UNIT,
	TTC_TRANSPONDER,
	IMU,
	SUN_SENSOR,
	STAR_SENSOR1,
	STAR_SENSOR2,
	GNC_COMPUTER,
	POWER_CONVERSION_AND_REGULATION_UNIT,
	BATTERY_HEATER,
	REACTOR_CONTROL_DRUMS,
	ENGINE_GIMBAL_DRIVE_A,
	ENGINE_GIMBAL_DRIVE_B,
	ENGINE_GIMBAL_DRIVE_C,
	TPA_A,
	TPA_B,
	TPA_C,
	CRYOCOOLER,
	JET_MIXER,  //?? should I include it with the pack?
	DOCKING_SYSTEM,
	CRYOSWITCH //???
};

#define MAX_CHARS_COMPONENT_NAME (30)

struct ComponentDesc_t
{
	char CompName[MAX_CHARS_COMPONENT_NAME];
	double MinPowerInput;
	double MaxPowerInput;
	double MinHeatLoad;
	double MaxHeatLoad;
	double MTTF; // mean time to failure, hours
};

typedef pair<double, string> command_t;

struct earlier : public binary_function<double, double, bool> {
	bool operator()(command_t x, command_t y) 
	{
		return (x.first > y.first); 
	};
};

//BUDGET:
//Sun Sensor
//triple or double star sensor
//imu
//GNC computer
//wiring losses
//power converter/regulator (peak-power tracker or direct?) PPT needed for cryocooler use?
//cryocooler
//heaters?
//s-band TT&C transceiver (nothing else needed?) dish with 4 dBi gain - 12W rx 40W tx??
//CD&H - 2W commands, 10W telemetry?
//RCS thruster consumption
//reactor control drums
//valves
//fuel sensors
//TPA circuitry
//engine/nozzle gimbal drives
//cryoswitch???

//RCS: N2O4+MMH: 1028.57 kg MMH 2571.43 kg N2O4 1/5 (Possibly) in forward tanks, 4/5s in the aft tanks, with cross feed to be sure.
// fnicking pressurant?

//Autopilots
const double NTR_PITCH_GIMBAL_LIMIT = 5.0; // 5 degrees max!
const double NTR_YAW_GIMBAL_LIMIT = 5.0; // 5 degrees max!

const double TVC_DAP_Gain = 5.0; //ok, it's free for now  but set to 3 to avoid overshoot?
const double RCS_DAP_Gain = 5.0; //not so free... set to 3 and it's fine?

//
const double DB_FINE = 0.05;
const double DB_MEDIUM = 0.5; //deg
const double DB_COARSE = 5.0; //deg

const double RATE_DB_DEFAULT = 0.0000001; //deg/s (?)

const double RATE_SLOWEST = 0.05; //deg/second
const double RATE_SLOW = 0.2;
const double RATE_MEDIUM = 0.5;
const double RATE_FAST = 2.0;

// this is some serious mess
const double DB_OFFSET = 0.044; //degrees
const double DB_SLOPE = -0.6; //deg/sec/degree
const double DB_HSLOPE = -0.8; //deg/sec/degree
const double DB_WL = 0.4; //deg/sec -from Apollo DAP
const double DB_RATE = 0.2; //deg/sec


const int NUM_CONTROLLERS = 5;

#define TVC_P (0)
#define TVC_Y (1)
#define RCS_P (2)
#define RCS_Y (3)
#define RCS_R (4)

//Self-destruct and transition criteria
const double LAUNCH_G_LIMIT = 7.0;
const int    LAUNCH_STATIC_dP_dt_STEP_LIMIT = 3;
const int    LAUNCH_G_STEP_LIMIT = 5;
const double LAUNCH_STATIC_dP_dt_THRESHOLD = 100.0;

const double MAX_DYN_PRESSURE = 1000; //1 kPa - don't aerobrake!

const double SAFE_PeA = 280.0;
const double SAFE_ApA = 420.0;

// Reactor physics
const double HALFLIFE_CONSTANT = 0.69314718;
const double StartupPeriod_s = 35; //conventional?
const double ShutdownPeriod_s = 70; //WTF??
const double Xe135_HalfLife_s = 9.1*3600; // Lange's, p.1.189
const double MaxAllowedXe135Poisoning = 0.05; // really?

//Physics in general
const double H2_CharacteristicVibrationT = 6159.0;
const double H2_RotationDOF = 2.0;
const double Boltzmann = 1.380622E-23;
const double H2_MolMass = 3.34E-27;
const double H2_R = Boltzmann/H2_MolMass;



//Animations
//const int NUM_ANIMATIONS = 2;
const int NUM_ANIMATIONS = 3;
#define SOLAR1_ANIMATION (0)
#define SOLAR2_ANIMATION (1)
#define SBAND_ANIMATION (2)

const VECTOR3 PVA1_Axis_Offset = _V( 3.19,-3.19,13.3); //13.45
const VECTOR3 PVA2_Axis_Offset = _V(-3.19, 3.19,13.3); //3.189052
const VECTOR3 SBand_Axis_Offset = _V(+3.273904, +3.273904,6.265);
const VECTOR3 DIR_XY_225_Z_0 = _V(-0.707107,-0.707107,0);
const VECTOR3 DIR_XY_045_Z_0 = _V(+0.707107,+0.707107,0);
const VECTOR3 DIR_XY_135_Z_0 = _V(-0.707107,+0.707107,0); //?

// animation transformation definitions
static UINT GRP_LWING = 2;
static UINT GRP_RWING = 3;


static VECTOR3 LWING_REF  = {-1.3,-0.725,-1.5};
static VECTOR3 LWING_AXIS = {-0.9619,-0.2735,0};
static VECTOR3 RWING_REF  = {1.3,-0.725,-1.5};
static VECTOR3 RWING_AXIS = {0.9619,-0.2735,0};
static float AILERON_RANGE = (float)(20.0*RAD);
static float ELEVATOR_RANGE = (float)(30.0*RAD);

//EXHAUSTS
const double  NTR_EXHAUST_LENGTH = 10.0;
const double  RCS_EXHAUST_LENGTH = 1.0;

//TODO: link with the inline tank and learn its fuel/RCS setup through the generic callback
// INHIBIT inline tank's RCS thrusters after docking!

// Grail-shaped beacon...
const double BEACON_OFFSET = 2.36086; //m, to coincide with the markings?
static VECTOR3 beaconPlusXpos = _V(BEACON_OFFSET, 0.0, 14.05);
static VECTOR3 beaconMinusXpos = _V(-BEACON_OFFSET, 0.0, 14.05);
static VECTOR3 beaconPlusYpos = _V(0.0, BEACON_OFFSET, 14.05);
static VECTOR3 beaconMinusYpos = _V(0.0, -BEACON_OFFSET, 14.05);
static VECTOR3 beaconColourRed = _V(1.0, 0.0, 0.0);
static VECTOR3 beaconColourGreen = _V(0.0, 1.0, 0.0);
static VECTOR3 beaconColourBlue = _V(0.0, 0.0, 1.0);
static VECTOR3 beaconColourWhite = _V(1.0, 1.0, 1.0);

const int NUM_BEACONS = 4;
#define BEACON_PLUSX  (0)
#define BEACON_MINUSX (1)
#define BEACON_PLUSY  (2)
#define BEACON_MINUSY (3)

// auxiliary classes
struct drawdown_t
{
	VESSEL * pVessel;
	int tankID;
};

enum reactor_state_t
{
	Offline = 0,
	UllageBurn,
	Startup,
	Nominal,
	Shutdown,
	Meltdown
};

#define ARE_PUMPS_WORKING(_StageState) (_StageState != reactor_state_t::Offline) \
	&& (_StageState != reactor_state_t::Meltdown) \
	&& (_StageState != reactor_state_t::UllageBurn)

#define ARE_PUMPS_NOT_WORKING(_StageState) (_StageState == reactor_state_t::Offline) \
	|| (_StageState == reactor_state_t::Meltdown) \
	|| (_StageState == reactor_state_t::UllageBurn)

enum turbopump_state_t
{
	off = 0,
	ok,
	xfa,
	xfb,
	xfc,
	out
};

enum simple_anim_status_t
{
	OPEN = 0,
	CLOSED,
	OPENING,
	CLOSING,
	LAST
};

enum glass_screen_t
{
	LAUNCH = 0,
	CHECKOUT,
	ENGINE_CONTROL,
	DAP_CONFIG,
	POWER_CONTROL,
	DOCKING,
	COMMUNICATIONS,
};

struct gimbal_problem_data_t
{
	int numThrusters;
	vector<VECTOR3> Pos;
	vector<double> Thrust;
	VECTOR3 COG;
};

enum PhasePlaneDecisionArea
{
	NO_CHANGE = 0,
	DA1,
	DA2,
	DA3,
	DA4,
	DA5,
	DA6
};

enum BatteryStatus_t
{
	CHARGED = 0,
	DISCHARGING,
	CHARGING,
	FAILED
};

inline double ComputeRateFromErrors(const double AttDB, const double AngRateError, const double AttError)
{
	// geometry:

	if (AngRateError>0.0)
	{
		
	}
	else
	{
	};
	return 0.0;
};

//CLASS INTERFACE
// ==============================================================
// NTR stage class interface
// ==============================================================

class NTR_test: public VESSEL3 {
public:
	NTR_test (OBJHANDLE hVessel, int flightmodel);
	~NTR_test ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void clbkPreStep (double simt, double simdt, double mjd);
	void clbkPostStep (double simt, double simdt, double mjd);

	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	void clbkSaveState  (FILEHANDLE  scn);
	int clbkConsumeBufferedKey  ( DWORD  key,  bool  down,  char *  kstate );
	void clbkDockEvent  ( int  dock,  OBJHANDLE  mate);
	void clbkPostCreation  (); 
	int clbkGeneric(int msgid, int prm, void *context);
	void clbkVisualCreated  ( VISHANDLE  vis,  int  refcount);
	void clbkVisualDestroyed  ( VISHANDLE  vis, int  refcount);

	bool GimbalNTR(int engine, double pitch, double yaw);//from SSU
	bool GimbalRollNTR(double R, double theta_deg);
	double GetStackMass();
	bool GetStackGroundContact();

	void SetRCSstringLevel_SingleStep(const stack_RCS_string_t& s, const double dlevel) const;
	void SetThrusterGroupLevel_SingleStep(THGROUP_TYPE  thgt,  double  dlevel) const;

	void ComputeOptimumRCS(double refThrust);

	void SelfDestruct(const char * reason);

	void DefineAnimations();

	double thermalPower, currentThrust, currentIsp, reactorPeriod_s; 
	//bool Startup, Shutdown, EngineOn, NominalPower, Meltdown;
	// valves? probabilities of failures?
	reactor_state_t StageState;

	bool MaintainCommandedRates;

	double dV_sinceBurnStart, massAtBurnStart;
	double XePoisoning;
	//INDIVIDUAL
	double engXePoisoning[NUM_REACTORS];

	vector<stackVessel_t> stackInfo;
	vector<drawdown_t> drawdownSequence;
	THRUSTER_HANDLE th_NTR[3];
	PROPELLANT_HANDLE hpr_LH2, hpr_RCS;
	double commanded_dV;
	double estimatedBurnTime_s;
	double NTRGimbal[NUM_REACTORS][2];
	bool GimbalSelect[NUM_REACTORS];
	double EngineThrust[NUM_REACTORS];
	//INDIVIDUAL
	double engCurrentIsp[NUM_REACTORS];
	double engReactorPeriod_s[NUM_REACTORS]; 
	double ReactorPower[NUM_REACTORS];
	reactor_state_t ReactorState[NUM_REACTORS];
	turbopump_state_t TPAState[NUM_REACTORS][2];
	PIDControl Controllers[NUM_CONTROLLERS]; //TVCControlP,TVCControlY, RCSControlR, RCSControlP, RCSControlY;
	bool FuelLevelTripActivated;
	bool AutoRCSEnabled;
	bool ShallRecomputeRCSStrings;
	double AutoRCSTorqueOrThrust;
	bool IsInLaunchMode;
	bool IsInCheckoutMode;
	int Static_dP_dt_criterion_counter;
	int G_limit_criterion_counter;
	double old_STP;
	double NTR_Life_s; // seconds
	double MissionTime_s; // seconds - for solar array output and other purposes

	simple_anim_status_t AnimationStatus[NUM_ANIMATIONS];
	static double AnimationSpeeds[NUM_ANIMATIONS];
	double AnimationDeployments[NUM_ANIMATIONS];
	static UINT Animations[NUM_ANIMATIONS];
	VECTOR3 CommandedRates;
	VECTOR3 Att_Deadbands; //TODO!
	VECTOR3 Rate_Deadbands;
	double DAP_SteadyRate; //degrees
	double DAP_SlowRate;
	// DONE: gains for the TVC and RCS controllers TODO: saving/reloading them
	bool ActiveDAPChannels[3];

	bool ShallTrackSun;
	bool SetRCSString(char *str);

	bool ShallTrackPoint;
	VECTOR3 TrackPointEulers;
	double UllageBurnTimeLeft;
	glass_screen_t CurrentScreen;
	stack_RCS_config_t stack_RCS;
	void OptimizeTorque(const vector<double> & thrust, vector<double> &pitch, vector<double> &yaw); //
	//double RateFromAttError(const double AttError, const double Deadband);
	double RateFromAttError(const double AttError, const double Deadband, const double RateError = 0.0);

	VECTOR3 RatesFromAttErrors(const VECTOR3 CurrentAttitude, const VECTOR3 DesiredAttitude, 
									 const VECTOR3 AttError,
									const VECTOR3 AttDeadbands, const VECTOR3 CurrentRates, 
									const bool IgnoreRoll = false);// rates in degrees?

	VECTOR3 stackCOG, stackPMI;
	vector<stack_thruster_t> stackThrusters;
	vector<stack_fuel_tank_t> stackTanks;
	SolSystem * solarSystem;

	BEACONLIGHTSPEC beacons[NUM_BEACONS];

	double SolarArrayPowerOutput;
	double BatteryPowerOutput;
	double NetPowerOutput;
	BatteryStatus_t BatteryStatus;
	double BatteryCharge; //Wh
	double BatteryCapacity; //Wh, degradable! (TBD)
	double BatteryActivationThreshold; //W
	int NumBatteryCycles;//not yet implemented - should be DOD dependent

	double TotalHeatLoad; //Watts!
	double T_ExternalSurface; //K
	double T_LH2_Tank; //K
	double CryocoolerPowerInput;
	double Boiloff_per_day; //kg
	hash_map<NTR_Component_t, bool> SystemHealthMap;

	UINT Exhaust_A, Exhaust_B, Exhaust_C;

	void Shutdown();
	void ToggleTrackSun();
	void TrackSun(const bool ShallTrack);
	//vector<command_t> CommandList;
	priority_queue<command_t, vector<command_t>, earlier> CommandList;
	void LoadCommands(const char * cmdFileName);
private:
	static void vlift (VESSEL *v, double aoa, double M, double Re,
		void *context, double *cl, double *cm, double *cd);
	static void hlift (VESSEL *v, double aoa, double M, double Re,
		void *context, double *cl, double *cm, double *cd);

	// transformations for control surface animations - there are no FCS here, but saved for safety anyway
	//static MGROUP_ROTATE trans_Laileron, trans_Raileron;
	//static MGROUP_ROTATE trans_Lelevator, trans_Relevator;


	void FillDrawdownSequence_and_OptimizeRCS();
	void FillDrawdownSequence();

	VISHANDLE visual;
	UINT MainMeshID;
	double fullCOG_elev;
	double emptyCOG_elev;
	bool shallFillDrawdownSeq;
	nlopt::opt* optProblem;
	float oldNozzleR, oldNozzleG, oldNozzleB;
	
	double timeEngineBurnStarted;
};

bool InputConfigParam (void *id, char *str, void *usrdata);

#endif //!__NTR_TEST_H__