#include "PIDControl.h"
//#include "SSUMath.h"
#include <string.h>

PIDControl::PIDControl(const char * Name, double pGain, double dGain, double iGain)
	: integral(0.0), lastValue(0.0), firstStep(true), outputLimitsSet(false), integralLimitsSet(false)
{
	strcpy_s(name, 19, Name);
	SetGains(pGain, dGain, iGain);
}

PIDControl::PIDControl(const char * Name, double pGain, double dGain, double iGain, double min, double max)
	: integral(0.0), lastValue(0.0), firstStep(true), integralLimitsSet(false)
{
	strcpy_s(name, 19, Name);
	SetGains(pGain, dGain, iGain);
	SetLimits(min, max);
}

PIDControl::PIDControl(const char * Name, double pGain, double dGain, double iGain, double min, double max, double iMin, double iMax)
	: integral(0.0), lastValue(0.0), firstStep(true)
{
	strcpy_s(name, 19, Name);
	SetGains(pGain, dGain, iGain);
	SetLimits(min, max);
	SetIntegralLimits(iMin, iMax);
}

PIDControl::~PIDControl()
{
}

void PIDControl::Reset()
{
	integral=0.0;
	lastValue=0.0;
	firstStep=true;
}

double PIDControl::Step(double value, double DeltaT)
{
	double derivative=0.0;
	
	if(!firstStep) {
		derivative = (value-lastValue)/DeltaT;
		//for the moment, use trapezoidal rule to calculate integral
		integral += DeltaT*((value+lastValue)/2.0);
	}
	else firstStep=false;
	//store value for next timestep
	lastValue=value;
	
	double limitedIntegral;
	if(integralLimitsSet)
		limitedIntegral = range(integralMin, integral, integralMax);
	else limitedIntegral = integral;

	double result = proportionalGain * value 
		+ derivativeGain * derivative
		+ integralGain * limitedIntegral;
	if(outputLimitsSet) result = range(minOutput, result, maxOutput);
	return result;
}

void PIDControl::SetGains(double pGain, double dGain, double iGain)
{
	proportionalGain=pGain;
	derivativeGain=dGain;
	integralGain=iGain;
}

void PIDControl::SetLimits(double min, double max)
{
	minOutput = min;
	maxOutput = max;
	outputLimitsSet = true;
}

void PIDControl::SetIntegralLimits(double min, double max)
{
	integralMin = min;
	integralMax = max;
	integralLimitsSet = true;
}