#ifndef __ECLIPSE_H_
#define __ECLIPSE_H_

#include "orbiterapi.h"
#include "solarsys.h"

// Wertz p.76
inline double ShadowConeLength(const OBJHANDLE hBody, SolSystem* solSys)
{
	OBJHANDLE hStar = GetCentralStar(solSys);
	
	//OBJHANDLE hStar = GetCentralStar();
	double R_s = oapiGetSize(hStar);
	double R_b = oapiGetSize(hBody);
	double S;
	VECTOR3 rpos;
	oapiGetRelativePos(hBody, hStar, &rpos);
	S = length(rpos);
	return (R_b*S)/(R_s-R_b);
}

inline double ShadowConeAngRadius_rad(const OBJHANDLE hBody, SolSystem* solSys)
{
	OBJHANDLE hStar = GetCentralStar(solSys);
	//OBJHANDLE hStar = GetCentralStar();
	double R_s = oapiGetSize(hStar);
	double R_b = oapiGetSize(hBody);
	double S;
	VECTOR3 rpos;
	oapiGetRelativePos(hBody, hStar, &rpos);
	S = length(rpos);
	return asin((R_s-R_b)/S);
}

inline void ShadowConeLengthAndRadius(const OBJHANDLE hBody, double& ConeLength, double& AngRadius_rad, 
									  SolSystem* solSys)
{
	OBJHANDLE hStar = GetCentralStar(solSys);
	double R_s = oapiGetSize(hStar);
	double R_b = oapiGetSize(hBody);
	double S;
	VECTOR3 rpos;
	oapiGetRelativePos(hBody, hStar, &rpos);
	S = length(rpos);
	ConeLength = (R_b*S)/(R_s-R_b);
	AngRadius_rad = asin((R_s-R_b)/S);
}

enum ECLIPSE {NO_ECLIPSE, PARTIAL_ECLIPSE, ANNULAR_ECLIPSE, TOTAL_ECLIPSE}; 

// no idea why I'm there that far and still get eclipses from Earth!
inline ECLIPSE EclipseState(const OBJHANDLE hVessel, const OBJHANDLE hBody, SolSystem* solSys)
{
	OBJHANDLE hStar = GetCentralStar(solSys);
	double R_s = oapiGetSize(hStar);
	double R_b = oapiGetSize(hBody);
	VECTOR3 rpos_body;
	oapiGetRelativePos(hBody, hStar, &rpos_body);
	double S = length(rpos_body);
	VECTOR3 rpos_v2b, rpos_v2s;
	oapiGetRelativePos(hVessel, hBody, &rpos_v2b);
	oapiGetRelativePos(hVessel, hStar, &rpos_v2s);
	double D_s = length(rpos_v2s);
	double D_b = length(rpos_v2b);
	double rho_S = asin(R_s/D_s);
	double rho_B = asin(R_b/D_b);
	// ang separation:
	//double theta = acos(dotp(-rpos_v2s, -rpos_v2b));
	double theta = angsep(-rpos_v2s, -rpos_v2b);
	double C = (R_b*S)/(R_s-R_b);
	//AngRadius_rad = asin((R_s-R_b)/S);
	if ((D_s>S) && (theta>fabs(rho_B-rho_S)) && ((rho_B+rho_S)>theta)) return PARTIAL_ECLIPSE; // this is what it consistently returns now.
	if ((S<D_s) && (D_s<(S+C)) && ((rho_B-rho_S)>theta)) return TOTAL_ECLIPSE;
	if (((S+C)<D_s) && ((rho_B-rho_S)>theta)) return ANNULAR_ECLIPSE;
	return NO_ECLIPSE;
}


//SpeechApi
// photosphere: radius is 6.9599E+5 km

//TODO: switch away from m to mu directly?

// doesn't work correctly during eclipses, where the heck is annular and partial eclipse?
// FIXED because of acos abomination, MUST LEARN GEOMETRY!
inline double IlluminationFactor(const OBJHANDLE hVessel, const OBJHANDLE hBody, SolSystem* solSys)
{
	ECLIPSE es = NO_ECLIPSE;
	OBJHANDLE hStar = GetCentralStar(solSys);
	double R_s = oapiGetSize(hStar);
	double R_b = oapiGetSize(hBody);
	VECTOR3 rpos_body;
	oapiGetRelativePos(hBody, hStar, &rpos_body);
	double S = length(rpos_body);
	VECTOR3 rpos_v2b, rpos_v2s;
	oapiGetRelativePos(hBody, hVessel, &rpos_v2b);
	oapiGetRelativePos(hStar, hVessel, &rpos_v2s);
	double D_s = length(rpos_v2s);
	double D_b = length(rpos_v2b);
	double rho_S = asin(R_s/D_s);
	double rho_B = asin(R_b/D_b);
	// ang separation:
	double theta = angsep(-rpos_v2s, -rpos_v2b);
	//double theta = acos(dotp(rpos_v2s, rpos_v2b));
	double C = (R_b*S)/(R_s-R_b);

	if ((D_s>S) && (theta>fabs(rho_B-rho_S)) && ((rho_B+rho_S)>theta)) 
	{
		es = PARTIAL_ECLIPSE;
	}
	else if ((S<D_s) && (D_s<(S+C)) && ((rho_B-rho_S)>theta)) 
	{
		es = TOTAL_ECLIPSE;
	}
	else if (((S+C)<D_s) && ((rho_B-rho_S)>theta)) 
	{
		es = ANNULAR_ECLIPSE;
	};
	double i0_minus_i;
	double cosRhoS, sinRhoS, cosRhoB, sinRhoB, sinTheta, cosTheta;
	switch (es)
	{
	case NO_ECLIPSE:
		return 1.0;
		break;
	case TOTAL_ECLIPSE:
		return 0.0;
		break;
	case PARTIAL_ECLIPSE:
		//3-59, 3-60

		cosRhoS = cos(rho_S);
		cosRhoB = cos(rho_B);
		sinRhoS = sin(rho_S);
		sinRhoB = sin(rho_B);
		sinTheta = sin(theta);		
		cosTheta = cos(theta);		
		i0_minus_i = (PI-cosRhoS*acos((cosRhoB-cosRhoS*cosTheta)/(sinRhoS*sinTheta))
			-cosRhoB*acos((cosRhoS-cosRhoB*cosTheta)/(sinRhoB*sinTheta))
			-acos((cosTheta-cosRhoB*cosRhoS)/(sinRhoB*sinRhoS)))/(PI*(1.0-cosRhoS));
		return 1.0-i0_minus_i;
		//return i0_minus_i;

		break;
	case ANNULAR_ECLIPSE:

		cosRhoS = cos(rho_S);
		cosRhoB = cos(rho_B);
		return 1.0-(1.0-cosRhoB)/(1.0-cosRhoS);
		//return (1.0-cosRhoB)/(1.0-cosRhoS);

		break;
	default:
		oapiWriteLog("Unknown eclipse state");
		throw "Unknown eclipse state";
	};
}

#endif //!__ECLIPSE_H_
