#ifndef __JACOBI_H_
#define __JACOBI_H_

#define STRICT

#include "optimization.h"
#include <math.h>
#include <limits>
#include <vector>
#include "orbitersdk.h"

using namespace std;
using namespace alglib;


typedef struct {
	double (*Function) (const real_1d_array &x, void* ptr_x); 
	void * Params;
	double eps;
} gradient_info_t;

void FiniteDifferenceGradient(const real_1d_array &x, double &func_val, real_1d_array &grad, void *ptr)
{
	// the bug was here...
	gradient_info_t* gt = (gradient_info_t*)ptr;
	double (*callback) (const real_1d_array &x, void* ptr_x) = gt->Function;
	void * data = gt->Params;
	double fval, fval_j, h, tmp;
	real_1d_array tempX;
	int N_in = x.length();
	tempX.setlength(N_in);

	for (int i = 0; i< N_in; i++) tempX[i] = x[i];
	fval = callback(tempX, data);

	for (int j = 0; j< N_in; j++)
	{
		tmp = tempX[j];
		h = gt->eps * fabs(tmp);
		if (fabs(h) < gt->eps) h = gt->eps;
		tempX[j] = tmp + h;
		fval_j = callback(tempX, data);
		tempX[j] = tmp;
		grad[j] = (fval_j - fval) / h;
	};//j
	//sprintf_s(???WTF???, 255, "x: %.2f %.2f %.2f f: %.2f df/dx: %.2f %.2f %.2f", x[0], x[1], x[2], fval, grad[0], grad[1], grad[2]);
	func_val = fval;
	//delete tempX;//??
}

// inverse kinematics: minimize weighted squared deviations from the target point

// x isn't passed
void minimize_constrained(const int N, const real_1d_array& x_initial, 
						  const real_1d_array& x_lower, 
						  const real_1d_array& x_upper, 
						  real_1d_array& x, double & min_func,
						  double (*Function) (const real_1d_array &x, void* ptr_x), 
						  void * Params = NULL, 
						  const double epsG = 0.0, 
						  const double epsF = 0.0, 
						  const double epsX = 0.0, 
						  const int maxIterations = 0)
{
	double epsg = epsG;
	double epsf = epsF;
	double epsx = epsX;
	double macheps = numeric_limits<double>::epsilon();
	int maxiterations = maxIterations;
	if (epsx == 0.0) epsx = sqrt(macheps);
	if (epsf == 0.0) epsf = sqrt(macheps);
	if (epsg == 0.0) epsg = sqrt(macheps);
	if (maxiterations==0) maxiterations = 100;
	gradient_info_t gt;
	gt.eps = 0.1; //?? m/s
	//gt.eps = epsg;
	gt.Params = Params;
	gt.Function = Function;
	minasastate state;
	minasacreate(N, x_initial, x_lower, x_upper, state);
	minasasetcond(state, epsg, epsf, epsx, maxiterations);
	minasasetxrep(state, false);
	minasasetalgorithm(state, -1);
	minasaoptimize(state, FiniteDifferenceGradient, NULL, &gt);
	minasareport rep;
	minasaresults(state, x, rep);
	
	//sprintf_s(oapiDebugString(), 255, "%d iterations; %d func evals", rep.iterationscount, rep.nfev);

	/*switch (rep.terminationtype)
	{
	case -2:
		sprintf_s(oapiDebugString(), 255, "rounding errors prevent further improvement. X contains best point found.");
		break;
	case -1:
        sprintf_s(oapiDebugString(), 255, "incorrect parameters were specified");
		break;
	case 0:
        sprintf_s(oapiDebugString(), 255, "ok?");
		break;
	case 1:
        sprintf_s(oapiDebugString(), 255, "relative function improvement is no more than EpsF");
		break;
	case 2:
        sprintf_s(oapiDebugString(), 255, "relative step is no more than EpsX.");
		break;
	case 4:
        sprintf_s(oapiDebugString(), 255, "gradient norm is no more than EpsG");
		break;
	case 5:
        sprintf_s(oapiDebugString(), 255, "maximum iterations limit reached");
		break;
	case 7:
        sprintf_s(oapiDebugString(), 255, "stopping conditions are too stringent, further improvement is impossible");
		break;
	default:
		break;
	};*/
	min_func = Function(x, Params);
	//sprintf_s(oapiDebugString(), 255, "%.2f %.2f %.2f", x[0], x[1], x[2]); // so the zeros are within the optimizer!
	
}

#endif //!__JACOBI_H_