#ifndef __PHYSCONSTANTS_H_
#define __PHYSCONSTANTS_H_

#include <math.h>

#define WGS84_A (6378137.0f)
#define Earth_J2_notide (0.00108262668f)
#define Earth_J2_tide (0.0010826359f)
// for Earth only!!
#define w_sun (360.0f / 365.242199f)

#define RAD_TO_DEG (57.2957795130823f)
#define DEG_TO_RAD (0.0174532925199433f)
// in Pa (N/m^2)
#define SolPressure_1AU (0.00000915f)
#define Stefan_Boltzmann_sigma (5.6704e-8)
#define SolarConstant_1AU (1370.0f)

#define J2000_Ecliptic_Obliquity (23.4392911f)
#define MJD_offset (2400000.5f)
#define HALFPI (1.5707963267949)
#define FORTPI (0.785398163397448)

inline int trunc2(const double v)
{
	double w, f;
	f = modf(v, &w);
	return (int) w;
};

//Credit: Andreas Horstemeier, Meeus (ch.7
inline void MJD_To_YMD(const double MJD, int& year, int& month, int& day)
{
  int z,a,b,c,d,e;
  double juldat = MJD+0.5+MJD_offset;
  z=trunc2(juldat);
  a=z;
  b=a+1524;
  c=trunc2((b-122.1)/365.25);
  d=trunc2(365.25*c);
  e=trunc2((b-d)/30.6001);
  day=b-d-trunc2(30.6001*e);
  year=c-4716;
  if (e<14)
  {
    month=e-1;
  }
  else
  {
    month=e-13;
    year=year+1;
  };
}

inline int MJD_To_Year(const double MJD)
{
	int y, m, d;
	MJD_To_YMD(MJD, y, m, d);
	return y;
}

#endif //!__PHYSCONSTANTS_H_