#ifndef __ROTATIONS_H_
#define __ROTATIONS_H_

#define STRICT

#include "SpiceUsr.h"
#include "orbitersdk.h"
#include <iostream>
#include "physconstants.h"

//#include <stdio.h>

#define EULER1 1
#define EULER2 2
#define EULER3 3

// Null vector
const VECTOR3 NULL_VECTOR = { 0.0, 0.0, 0.0 };

enum AXIS {PITCH, YAW, ROLL};

enum ROT_AXIS {AXIS_X, AXIS_Y, AXIS_Z};

enum AXISASSOCIATION
{
	PLUS_X, MINUS_X, PLUS_Y, MINUS_Y, PLUS_Z, MINUS_Z
};

typedef struct {
	VECTOR3 Pitch;
	VECTOR3 Yaw;
	VECTOR3 Roll;			
} RefPoints;

using namespace std;

// Are the vectors equal?
inline double operator== (const VECTOR3 &a, const VECTOR3 &b)
{
	return ((a.x == b.x) && (a.y == b.y) && (a.z == b.z));
}

// pre-mul is missing from orbitersdk.
inline VECTOR3 operator* (const double s, const VECTOR3 &a)
{
	return _V(s*a.x, s*a.y, s*a.z);
}

inline MATRIX3 operator* (const double s, const MATRIX3 &m)
{
	return m*s;
}

inline VECTOR3 ForceLocalToGlobal(const OBJHANDLE hObj, const VECTOR3 vec)
{
	MATRIX3 m;
	oapiGetRotationMatrix(hObj, &m);
	return tmul(m, vec);
}

inline VECTOR3 ForceGlobalToLocal(const OBJHANDLE hObj, const VECTOR3 vec)
{
	MATRIX3 m;
	oapiGetRotationMatrix(hObj, &m);
	return mul(m, vec);
}

// Returns the rotation matrix for a rotation of a given angle around the X axis (Pitch)
void inline GetRotMatrixX(double Angle, MATRIX3 &RotMatrixX)
{
	RotMatrixX.m11 = 1;
	RotMatrixX.m12 = 0;
	RotMatrixX.m13 = 0;
	RotMatrixX.m21 = 0;
	RotMatrixX.m22 = cos(Angle);
	RotMatrixX.m23 = sin(Angle);
	RotMatrixX.m31 = 0;
	RotMatrixX.m32 = -sin(Angle);
	RotMatrixX.m33 = cos(Angle);
}

inline MATRIX3 GetRotMatrixX_dcm(double cosAngle, int side)
{
	double sinAngle = side*sqrt(1-cosAngle*cosAngle);
	MATRIX3 RotMatrixX;
	RotMatrixX.m11 = 1;
	RotMatrixX.m12 = 0;
	RotMatrixX.m13 = 0;
	RotMatrixX.m21 = 0;
	RotMatrixX.m22 = cosAngle;
	RotMatrixX.m23 = sinAngle;
	RotMatrixX.m31 = 0;
	RotMatrixX.m32 = -sinAngle;
	RotMatrixX.m33 = cosAngle;
	return RotMatrixX;
}

// Returns the rotation matrix for a rotation of a given angle around the Y axis (Yaw)
void inline GetRotMatrixY(double Angle, MATRIX3 &RotMatrixY)
{
	RotMatrixY.m11 = cos(Angle);
	RotMatrixY.m12 = 0;
	RotMatrixY.m13 = -sin(Angle);
	RotMatrixY.m21 = 0;
	RotMatrixY.m22 = 1;
	RotMatrixY.m23 = 0;
	RotMatrixY.m31 = sin(Angle);
	RotMatrixY.m32 = 0;
	RotMatrixY.m33 = cos(Angle);
}

inline MATRIX3 GetRotMatrixY_dcm(double cosAngle, int side)
{
	double sinAngle = side*sqrt(1-cosAngle*cosAngle);
	MATRIX3 RotMatrixY;
	RotMatrixY.m11 = cosAngle;
	RotMatrixY.m12 = 0;
	RotMatrixY.m13 = -sinAngle;
	RotMatrixY.m21 = 0;
	RotMatrixY.m22 = 1;
	RotMatrixY.m23 = 0;
	RotMatrixY.m31 = sinAngle;
	RotMatrixY.m32 = 0;
	RotMatrixY.m33 = cosAngle;
	return RotMatrixY;
}

// Returns the rotation matrix for a rotation of a given angle around the Z axis (Roll)
void inline GetRotMatrixZ(double Angle, MATRIX3 &RotMatrixZ)
{
	RotMatrixZ.m11 = cos(Angle);
	RotMatrixZ.m12 = sin(Angle);
	RotMatrixZ.m13 = 0;
	RotMatrixZ.m21 = -sin(Angle);
	RotMatrixZ.m22 = cos(Angle);
	RotMatrixZ.m23 = 0;
	RotMatrixZ.m31 = 0;
	RotMatrixZ.m32 = 0;
	RotMatrixZ.m33 = 1;
}

inline MATRIX3 GetRotMatrixZ_dcm(double cosAngle, int side)
{
	double sinAngle = side*sqrt(1-cosAngle*cosAngle);
	MATRIX3 RotMatrixZ;
	RotMatrixZ.m11 = cosAngle;
	RotMatrixZ.m12 = sinAngle;
	RotMatrixZ.m13 = 0;
	RotMatrixZ.m21 = -sinAngle;
	RotMatrixZ.m22 = cosAngle;
	RotMatrixZ.m23 = 0;
	RotMatrixZ.m31 = 0;
	RotMatrixZ.m32 = 0;
	RotMatrixZ.m33 = 1;
	return RotMatrixZ;
}

inline MATRIX3 GetRotMatrix(const VECTOR3 EulerAngles)
{
	MATRIX3 rx, ry, rz;
	GetRotMatrixX(EulerAngles.data[PITCH], rx);
	GetRotMatrixY(EulerAngles.data[YAW], ry);
	GetRotMatrixZ(EulerAngles.data[ROLL], rz);
	MATRIX3 r1 = mul(rx, ry);
	MATRIX3 _out = mul(r1, rz);
	return _out;
}

#define EulerToMatrix(ANG) GetRotMatrix(ANG)

inline MATRIX3 GetRotMatrix_dcm(const VECTOR3 cosAngles, const int cube)
{
	MATRIX3 rx, ry, rz;
	int pitchside, yawside, rollside;
	switch (cube)
	{
	case 0:
		pitchside = yawside = rollside = 1;
		break;
	case 1:
		pitchside = yawside = 1;
		rollside = -1;
		break;
	case 2:
		pitchside = rollside = 1;
		yawside = -1;
		break;
	case 3:
		pitchside = 1;
		yawside = rollside = -1;
		break;
	case 4:
		pitchside = -1;
		yawside = rollside = 1;
		break;
	case 5:
		pitchside = rollside = -1;
		yawside = 1;
		break;
	case 6:
		pitchside = yawside = -1;
		rollside = 1;
		break;
	case 7:
		pitchside = yawside = rollside = -1;
		break;
	default:
		throw "invalid cube";
		break;
	};
	rx = GetRotMatrixX_dcm(cosAngles.data[PITCH], pitchside);
	ry = GetRotMatrixY_dcm(cosAngles.data[YAW], yawside);
	rz = GetRotMatrixZ_dcm(cosAngles.data[ROLL], rollside);
	MATRIX3 r1 = mul(rx, ry);
	MATRIX3 _out = mul(r1, rz);
	return _out;
}

void MultiplyByMatrix(const VECTOR3 &Initial, const MATRIX3 &RotMatrix, VECTOR3 &Result)
{

	Result.x =	(Initial.x * RotMatrix.m11) 
		+ (Initial.y * RotMatrix.m12) 
		+ (Initial.z * RotMatrix.m13);

	Result.y =	(Initial.x * RotMatrix.m21) 
		+ (Initial.y * RotMatrix.m22) 
		+ (Initial.z * RotMatrix.m23);

	Result.z =	(Initial.x * RotMatrix.m31) 
		+ (Initial.y * RotMatrix.m32) 
		+ (Initial.z * RotMatrix.m33);

}

void RotateVector(const VECTOR3 &Initial, const VECTOR3 &Angles, VECTOR3 &Result)
{
	MATRIX3 RotMatrixX, RotMatrixY, RotMatrixZ;
	VECTOR3 AfterZ, AfterZY;					// Temporary variables


	GetRotMatrixX(Angles.x, RotMatrixX);
	GetRotMatrixY(Angles.y, RotMatrixY);
	GetRotMatrixZ(Angles.z, RotMatrixZ);

	MultiplyByMatrix(Initial, RotMatrixZ, AfterZ);
	MultiplyByMatrix(AfterZ, RotMatrixY, AfterZY);
	MultiplyByMatrix(AfterZY, RotMatrixX, Result);
}

// Return a vector containing the pitch, yaw, and roll angles between the current spacecraft
// attitude and the attitude defined by the two vectors; the first defines the pitch and yaw,
// and the second defines the roll.  If the second vector is the null vector, the function
// returns only the pitch and yaw (roll is assigned 0.0).  The input vectors should be with
// respect to the local frame
VECTOR3 GetPitchYawRoll(VECTOR3 PitchYaw, VECTOR3 Roll)
{
	double Distance;
	VECTOR3 Res;

	Distance = sqrt(pow(PitchYaw.x, 2) + pow(PitchYaw.y, 2) + pow(PitchYaw.z, 2));
	Res.data[PITCH] = asin(PitchYaw.y/Distance);

	if (PitchYaw.z < 0) {
		//Res.data[PITCH] = -Res.data[PITCH];

		//Res.data[PITCH] = PI - Res.data[PITCH];
	}


	// Yaw 	
	Res.data[YAW] = acos(PitchYaw.z/(Distance * sin(acos(PitchYaw.y/Distance))));

	//if (Res.data[PITCH] > PI/2) {
	//	Res.data[YAW] = PI - Res.data[YAW];	
	//}

	if (PitchYaw.x < 0) {
		Res.data[YAW] = -Res.data[YAW];
	}

	// Roll, if the second vector is not null
	if (Roll == NULL_VECTOR) {
		Res.data[ROLL] = 0.0;
		return Res;
	}

	Distance = sqrt(pow(Roll.x, 2) + pow(Roll.y, 2) + pow(Roll.z, 2));


	Res.data[ROLL] = asin(Roll.y / Distance);

	//if (Roll.z < 0) {
	//	Res.data[ROLL] = PI-Res.data[ROLL];
	//}

	return Res;
}

VECTOR3 inline GetPYR(VECTOR3 Pitch, VECTOR3 YawRoll)
{	
	VECTOR3 Res = { 0, 0, 0 };

	// Normalize the vectors
	VECTOR3 P = unit(Pitch);
	VECTOR3 YR = unit(YawRoll);
	VECTOR3 H = unit(crossp(P, YR));

	Res.data[YAW] = asin(P.x);

	Res.data[ROLL] = -atan2(H.x, YR.x);

	Res.data[PITCH] = -atan2(P.y, P.z);

	return Res;

}

// CODE FROM NASA WORLD WIND
struct quaternion_t 
{
public:
	union
	{
		double data[4];
		struct {
			double W;
			double X;
			double Y;
			double Z;
		};
	};

	quaternion_t() {W=0.0;X=0.0;Y=0.0;Z=1.0;}

	quaternion_t(const double w, const double x, const double y, const double z)
	{
		W = w;
		X = x;
		Y = y;
		Z = z;
	}

	/*theta  in radians*/

	quaternion_t(const double theta, const VECTOR3& axis)
	{
		double sinTheta = sin(theta/2.0);
		W=cos(theta/2.0);
		X=axis.x*sinTheta;
		Y=axis.y*sinTheta;
		Z=axis.z*sinTheta;
	}

	quaternion_t(const VECTOR3& vector, const double scalar)
	{
		W=scalar;
		X=vector.x;
		Y=vector.y;
		Z=vector.z;
	}

	inline quaternion_t& operator=(const quaternion_t& RHS)
	{
		W = RHS.W;
		X = RHS.X;
		Y = RHS.Y;
		Z = RHS.Z;
		return *this;
	}


	// conjugate operator
	inline quaternion_t Conjugate() const
	{
		quaternion_t qq;
		qq.X = -X;
		qq.Y = -Y;
		qq.Z = -Z;
		qq.W = W;
		return qq;
	}

	inline void norm()
	{
		double L = length();
		if (fabs(L)<std::numeric_limits<double>::epsilon()) throw "zero quaternion!";
		X = X / L;
		Y = Y / L;
		Z = Z / L;
		W = W / L;
	}

	inline VECTOR3 vector() const
	{
		return _V(X, Y, Z);
	}
	// what is the multiplicand that sets w to zero? ambiguity?
	// currently buggy.
	//inline quaternion_t Derotate() const
	//{
	//	quaternion_t q;
	//	double theta_2;
	//	theta_2 = acos(W);
	//	double sinTheta_2 = sin(theta_2);
	//	if (fabs(sinTheta_2)<std::numeric_limits<double>::epsilon()) throw "cannot perform derotation";
	//	q.W = 0.0;
	//	q.X = X/sinTheta_2;
	//	q.Y = Y/sinTheta_2;
	//	q.Z = Z/sinTheta_2;
	//	return q;
	//}

	inline quaternion_t unit() const
	{
		quaternion_t q;
		double L = length();
		if (fabs(L)<std::numeric_limits<double>::epsilon()) throw "zero quaternion!";
		q.X = X / L;
		q.Y = Y / L;
		q.Z = Z / L;
		q.W = W / L;
		return q;
	}
	//quaternion_t Ln() 
	//{
	//	return Ln(this);
	//}

	inline double length() const
	{
		return sqrt(X * X + Y * Y +
			Z * Z + W * W);
	}


	inline quaternion_t Inverse() const
	{
		quaternion_t qq;
		double mag_sq = W*W+X*X+Y*Y+Z*Z;
		if (fabs(mag_sq)<std::numeric_limits<double>::epsilon()) throw "zero quaternion!";
		qq.W = W/mag_sq;
		qq.X = -X/mag_sq;
		qq.Y = -Y/mag_sq;
		qq.Z = -Z/mag_sq;
		return qq;
	}

    friend ostream& operator<<(ostream& stream, const quaternion_t& operand);
    //friend istream& operator>>(istream& stream, quaternion_t& operand); // no input for now

	friend bool operator==(const quaternion_t& lhs, const quaternion_t& rhs);
	friend bool operator!=(const quaternion_t& lhs, const quaternion_t& rhs);
	friend double dotp(const quaternion_t& lhs, const quaternion_t& rhs);
	friend quaternion_t operator+(const quaternion_t a, const quaternion_t b);


};

// OK.

ostream& operator<<(ostream& stream, const quaternion_t& operand)
{
    //int prec = stream.Prec;
	//char buf[256];
	stream << "w:" << operand.W << " x:" << operand.X << " y:" << operand.Y << " z:" << operand.Z;
    return stream;
};

// OK
// Euler angles in radians
inline quaternion_t EulerToQuaternion(const VECTOR3 Euler)
{
	quaternion_t result;
	double mm[3][3];
	eul2m_c(Euler.z, Euler.y, Euler.x, EULER3, EULER2, EULER1, mm); // check order!
	double qq[4];
	m2q_c(mm, qq);
	result.W = qq[0];
	result.X = qq[1];
	result.Y = qq[2];
	result.Z = qq[3];
	return result;
	//double pitch= Euler.x;
	//double yaw = Euler.y;
	//double roll = Euler.z;
	//double cy = cos(yaw * 0.5);
	//double cp = cos(pitch * 0.5);
	//double cr = cos(roll * 0.5);
	//double sy = sin(yaw * 0.5);
	//double sp = sin(pitch * 0.5);
	//double sr = sin(roll * 0.5);

	
	///*double qw = cy*cp*cr + sy*sp*sr;
	//double qx = sy*cp*cr - cy*sp*sr;
	//double qy = cy*sp*cr + sy*cp*sr;
	//double qz = cy*cp*sr - sy*sp*cr;*/

	//result.W = cy*cp*cr + sy*sp*sr;
	//result.X = cy*sp*cr + sy*cp*sr; //p
	//result.Y = cy*sp*cr + sy*cp*sr; //p
	////result.Y = sy*cp*cr - cy*sp*sr; //y
	//result.X = sy*cp*cr - cy*sp*sr; //y
	//result.Z = cy*cp*sr - sy*sp*cr; //r
	//return result;
	// DOES NOT MAINTAIN UNIT NORM! WRONG!!!
}



inline VECTOR3 QuaternionToEuler(const quaternion_t q)
{
	VECTOR3 Euler;
	double qq[4];
	qq[0] = q.W;
	qq[1] = q.X;
	qq[2] = q.Y;
	qq[3] = q.Z;
	double mm[3][3];
	q2m_c(qq, mm);
	m2eul_c(mm, EULER3, EULER2, EULER1, &Euler.z, &Euler.y, &Euler.x); // check axis order!
	//	m2eul_c(mm, 1, 2, 3, &Euler.x, &Euler.y, &Euler.z); // check axis order!
	//double L = q.length(); // debug code? better be safe
	//if (fabs(L)<std::numeric_limits<double>::epsilon()) throw "quaternion has zero norm";
	//if (fabs(L-1.0)>std::numeric_limits<double>::epsilon()) throw "quaternion is not normalized";
	//double q0 = q.W;
	//double q1 = q.X;
	//double q2 = q.Y;
	//double q3 = q.Z;
	//Euler.x = asin( -2 * (q1*q3 - q0*q2)); //pitch??
	////Euler.y = asin( -2 * (q1*q3 - q0*q2)); //pitch??
	//Euler.y = atan2( 2 * (q2*q3 + q0*q1), (q0*q0 - q1*q1 - q2*q2 + q3*q3)); //yaw?
	////Euler.x = atan2( 2 * (q2*q3 + q0*q1), (q0*q0 - q1*q1 - q2*q2 + q3*q3)); //yaw?
	//Euler.z = atan2( 2 * (q1*q2 + q0*q3), (q0*q0 + q1*q1 - q2*q2 - q3*q3)); //roll
	return Euler;
}

inline MATRIX3 QuaternionToMatrix(const quaternion_t q)
{
	//Wertz Appx E.
	double w2 = q.W*q.W;
	double x2 = q.X*q.X;
	double y2 = q.Y*q.Y;
	double z2 = q.Z*q.Z;
	double q0q1 = q.W*q.X;
	double q1q2 = q.X*q.Y;
	double q2q3 = q.Y*q.Z;
	double q0q2 = q.W*q.Y;
	double q0q3 = q.W*q.Z;
	double q1q3 = q.X*q.Z;
	MATRIX3 m = _M(w2+x2-y2-z2,     2.0*(q1q2-q0q3), 2.0*(q1q3+q0q2),
		           2.0*(q1q2+q0q3), w2-x2+y2-z2,     2.0*(q2q3-q0q1),
				   2.0*(q1q3-q0q2), 2.0*(q2q3-q0q1), w2-x2-y2+z2); // ACK!!
					//  ^^                   ^
	return m;
}

inline MATRIX3 GibbsToMatrix(const VECTOR3 g)
{
	double s1 = g.x*g.x;
	double s2 = g.y*g.y;
	double s3 = g.z*g.z;
	double premul = 1.0/(1.0+s1+s2+s3);
	double g12 = g.x*g.y;
	double g13 = g.x*g.z;
	double g23 = g.y*g.z;
	MATRIX3 m = _M(1.0+s1-s2-s3, 2.0*(g12+g.z), 2.0*(g13-g.y),
				   2.0*(g12-g.z),1.0-s1+s2-s3,  2.0*(g23+g.x),
				   2.0*(g13+g.y),2.0*(g23-g.x),1.0-s1-s2+s3);
	return m*premul;
}

//some code was taken from Dragonfly's quaternion.cpp and rejected
//Ken Shoemake SIGGRAPH 85 Animating Rotation with Quaternion Curves V.19 No.3 1985. Pp.245-254
// Zipfel p.127 (4.80)
//
//inline MATRIX3 QuaternionToMatrix(const quaternion_t q)
//{
//	//from Zipfel. needs checking as to what translates into what... OMG...
//	/*double w2 = q.W*q.W;
//	double x2 = q.X*q.X;
//	double y2 = q.Y*q.Y;
//	double z2 = q.Z*q.Z;
//	double q0q1 = q.W*q.X;
//	double q1q2 = q.X*q.Y;
//	double q2q3 = q.Y*q.Z;
//	double q0q2 = q.W*q.Y;
//	double q0q3 = q.W*q.Z;
//	double q1q3 = q.X*q.Z;
//
//	MATRIX3 m = _M(w2+x2-y2-z2,     2.0*(q1q2-q0q3), 2.0*(q1q3+q0q2),
//		           2.0*(q1q2+q0q3), w2-x2+y2-z2,     2.0*(q2q3-q0q1),
//				   2.0*(q1q2-q0q2), 2.0*(q2q3+q0q1), w2-x2-y2+z2);*/
//	
//	MATRIX3 m;
//	//double m2[3][3];
//	//q2m_c(q.data, m2);
//	////could be wrong for all I know
//	//m = _M(m2[0][0], m2[0][1],m2[0][2],
//	//	   m2[1][0], m2[1][1],m2[1][2],
//	//	   m2[2][0], m2[2][1],m2[2][2]);
//	//return m;
//	VECTOR3 Eulers = QuaternionToEuler(q);
//	m = GetRotMatrix(Eulers);
//	return m;
//}

//p.253
inline quaternion_t MatrixToQuaternion(const MATRIX3 m)
{
	quaternion_t q;
	m2q_c(m.data, q.data);
	return q;
	double x_sq, y_sq, w_sq, four_w;
	w_sq = 0.25*(1.0+m.m11+m.m22+m.m33);
	if (w_sq>std::numeric_limits<double>::epsilon())
	{
		q.W = sqrt(w_sq);
		four_w = q.W*4.0;
		q.X = (m.m23-m.m32)/four_w;
		q.Y = (m.m31-m.m13)/four_w;
		q.Z = (m.m12-m.m21)/four_w;
	}
	else
	{
		q.W = 0.0;
		x_sq = -0.5*(m.m22+m.m33);
		if (x_sq>std::numeric_limits<double>::epsilon())
		{
			q.X = sqrt(x_sq);
			q.Y = m.m12/(2.0*q.X);
			q.Z = m.m13/(2.0*q.X);
		}
		else
		{
			q.X = 0.0;
			y_sq = 0.5*(1-m.m33);
			if (y_sq>std::numeric_limits<double>::epsilon())
			{
				q.Y = sqrt(y_sq);
				q.Z = m.m23/(2.0*q.Y);
			}
			else
			{
				q.Y = 0.0;
				q.Z = 1.0;
			};
		};
	};
	return q;
}

inline quaternion_t operator+(const quaternion_t a, const quaternion_t b)
{
	quaternion_t q;
	q.X = a.X + b.X;
	q.Y = a.Y + b.Y;
	q.Z = a.Z + b.Z;
	q.W = a.W + b.W;
	return q;
}

inline quaternion_t operator-(const quaternion_t a, const quaternion_t b)
{
	quaternion_t q;
	q.X = a.X - b.X;
	q.Y = a.Y - b.Y;
	q.Z = a.Z - b.Z;
	q.W = a.W - b.W;
	return q;
}

inline quaternion_t operator*(const quaternion_t a, const quaternion_t b)
{
	quaternion_t q;
	q.X = a.W * b.X + a.X * b.W + a.Y * b.Z - a.Z * b.Y;
	q.Y = a.W * b.Y + a.Y * b.W + a.Z * b.X - a.X * b.Z;
	q.Z = a.W * b.Z + a.Z * b.W + a.X * b.Y - a.Y * b.X;
	q.W = a.W * b.W - a.X * b.X - a.Y * b.Y - a.Z * b.Z;
	return q;
}

// pre-mul is equal to post-mul, for symmetry reasons

inline quaternion_t operator*(const double s, const quaternion_t q)
{
	quaternion_t qq;
	qq.X = s * q.X;
	qq.Y = s * q.Y;
	qq.Z = s * q.Z;
	qq.W = s * q.W;
	return qq;
}

inline quaternion_t operator*(const quaternion_t q, const double s)
{
	quaternion_t qq;
	qq.X = s * q.X;
	qq.Y = s * q.Y;
	qq.Z = s * q.Z;
	qq.W = s * q.W;
	return qq;
}

// equivalent to multiplying by the quaternion (0, v)
inline quaternion_t operator*(const VECTOR3 v, const quaternion_t q)
{
	quaternion_t qq;

	qq.X = v.x * q.W + v.y * q.Z - v.z * q.Y;
	qq.Y = v.y * q.W + v.z * q.X - v.x * q.Z;
	qq.Z = v.z * q.W + v.x * q.Y - v.y * q.X;
	qq.W = -v.x * q.X - v.y * q.Y - v.z * q.Z;
	return qq;
}

inline quaternion_t operator/(const quaternion_t q, const double s)
{
	return q * (1 / s);
}


inline double norm2(const quaternion_t& q)
{
	return q.X * q.X + q.Y * q.Y + q.Z * q.Z + q.W * q.W;
}

inline double abs(quaternion_t q)
{
	return sqrt(norm2(q));
}

inline quaternion_t operator/(const quaternion_t lhs, const quaternion_t rhs)
{
	return lhs * (rhs.Conjugate() / abs(rhs));
}

inline bool operator==(const quaternion_t& lhs, const quaternion_t& rhs)
{
	return lhs.X == rhs.X && lhs.Y == rhs.Y && lhs.Z == rhs.Z && lhs.W == rhs.W;
}

inline bool operator!=(const quaternion_t& lhs, const quaternion_t& rhs)
{
	return lhs.X != rhs.X || lhs.Y != rhs.Y || lhs.Z != rhs.Z || lhs.W != rhs.W;
}

inline double dotp(const quaternion_t& lhs, const quaternion_t& rhs)
{
	return lhs.X * rhs.X + lhs.Y * rhs.Y + lhs.Z * rhs.Z + lhs.W * rhs.W;
}

inline quaternion_t _Q1(const double w, const double x, const double y, const double z)
{
	quaternion_t _out;
	_out.W = w;
	_out.X = x;
	_out.Y = y;
	_out.Z = z;
	_out.norm();
	return _out;
}

inline quaternion_t _Q1(const double s, const VECTOR3 v)
{
	quaternion_t _out;
	_out.W = s;
	_out.X = v.x;
	_out.Y = v.y;
	_out.Z = v.z;
	_out.norm();
	return _out;
}

inline quaternion_t Slerp(const quaternion_t q0, const quaternion_t q1, double t)
{
	double cosom = q0.X * q1.X + q0.Y * q1.Y + q0.Z * q1.Z + q0.W * q1.W;
	double tmp0, tmp1, tmp2, tmp3;
	if (cosom < 0.0)
	{
		cosom = -cosom;
		tmp0 = -q1.X;
		tmp1 = -q1.Y;
		tmp2 = -q1.Z;
		tmp3 = -q1.W;
	}
	else
	{
		tmp0 = q1.X;
		tmp1 = q1.Y;
		tmp2 = q1.Z;
		tmp3 = q1.W;
	}

	/* calc coeffs */
	double scale0, scale1;

	if ((1.0 - cosom) > std::numeric_limits<double>::epsilon())
	{
		// standard case (slerp)
		double omega =  acos (cosom);
		double sinom = sin (omega);
		scale0 =  sin ((1.0 - t) * omega) / sinom;
		scale1 =  sin (t * omega) / sinom;
	}
	else
	{
		/* just lerp */
		scale0 = 1.0 - t;
		scale1 = t;
	}

	quaternion_t q;

	q.X = scale0 * q0.X + scale1 * tmp0;
	q.Y = scale0 * q0.Y + scale1 * tmp1;
	q.Z = scale0 * q0.Z + scale1 * tmp2;
	q.W = scale0 * q0.W + scale1 * tmp3;

	return q;
}

inline quaternion_t Ln(quaternion_t q)
{
	quaternion_t qq;
	double t = 0;

	double s = sqrt(q.X * q.X + q.Y * q.Y + q.Z * q.Z); 
	double om = atan2(s, q.W); 

	if (abs(s) < std::numeric_limits<double>::epsilon()) 
		t = 0.0f; 
	else 
		t = om/s; 

	qq.X = q.X * t;
	qq.Y = q.Y * t;
	qq.Z = q.Z * t;
	qq.W = 0.0f;

	return qq;
}

//the below functions have not been certified to work properly
inline quaternion_t Exp(quaternion_t q)
{
	quaternion_t qq;
	double sinom;
	double om = sqrt(q.X * q.X + q.Y * q.Y + q.Z * q.Z);

	if (abs(om) < std::numeric_limits<double>::epsilon()) 
		sinom = 1.0; 
	else 
		sinom = sin(om)/om; 

	qq.X = q.X * sinom;
	qq.Y = q.Y * sinom;
	qq.Z = q.Z * sinom;
	qq.W = cos(om);

	return qq;
}

inline quaternion_t Squad(
						  quaternion_t q1,
						  quaternion_t a,
						  quaternion_t b,
						  quaternion_t c,
						  double t)
{
	return Slerp(
		Slerp(q1, c, t), Slerp(a, b, t), 2 * t * (1.0 - t));
}

void SquadSetup(
				quaternion_t& outA,
				quaternion_t& outB,
				quaternion_t& outC,
				quaternion_t q0,
				quaternion_t q1,
				quaternion_t q2,
				quaternion_t q3)
{
	q0 = q0 + q1;
	q0.norm();

	q2 = q2 + q1;
	q2.norm();

	q3 = q3 + q1;
	q3.norm();

	q1.norm();

	outA = q1 * Exp(-0.25 * (Ln(Exp(q1) * q2) + Ln(Exp(q1) * q0)));
	outB = q2 * Exp(-0.25 * (Ln(Exp(q2) * q3) + Ln(Exp(q2) * q1)));
	outC = q2;
}



inline MATRIX3 TwoVectorFrame(const VECTOR3& PRI, 
							  const AXISASSOCIATION PrimaryAxisAssociation, 
							  const VECTOR3& SEC,
							  const AXISASSOCIATION SecondaryAxisAssociation)
{
	//VECTOR3 PRI, SEC;
	VECTOR3 X, Y, Z;
	/*double separation = fabs(acos(dotp(PRI, SEC)));
	if (separation<0.001) 
	{
	oapiWriteLog("Angle separation between primary and secondary vectors is less than 1 mrad");
	throw "Angle separation between primary and secondary vectors is less than 1 mrad";
	};*/
	switch (PrimaryAxisAssociation)
	{
	case PLUS_X:
	case MINUS_X:
		X = (PrimaryAxisAssociation==PLUS_X ? unit(PRI) : -unit(PRI));
		switch (SecondaryAxisAssociation)
		{
		case PLUS_Y:
		case MINUS_Y:
			Y = (SecondaryAxisAssociation==PLUS_Y ? unit(crossp(X, SEC)) : -unit(crossp(X, SEC)));
			Z = crossp(X, Y);
			break;
		case PLUS_Z:
		case MINUS_Z:
			Z = (SecondaryAxisAssociation==PLUS_Z ? unit(crossp(X, SEC)) : -unit(crossp(X, SEC)));
			Y = crossp(Z, X);
			break;
		default:
			oapiWriteLog("unknown or invalid secondary axis association");
			throw "unknown or invalid secondary axis association";
			break;
		};
		break;
	case PLUS_Y:
	case MINUS_Y:
		Y = (PrimaryAxisAssociation==PLUS_Y ? unit(PRI) : -unit(PRI));
		switch (SecondaryAxisAssociation)
		{
		case PLUS_X:
		case MINUS_X:
			X = (SecondaryAxisAssociation==PLUS_X ? unit(crossp(Y, SEC)) : -unit(crossp(Y, SEC)));
			Z = crossp(X, Y);
			break;
		case PLUS_Z:
		case MINUS_Z:
			Z = (SecondaryAxisAssociation==PLUS_Z ? unit(crossp(X, SEC)) : -unit(crossp(X, SEC)));
			X = crossp(Y, Z);
			break;
		default:
			oapiWriteLog("unknown or invalid secondary axis association");
			throw "unknown or invalid secondary axis association";
			break;
		};
		break;
	case PLUS_Z:
	case MINUS_Z:
		Z = (PrimaryAxisAssociation==PLUS_Z ? unit(PRI) : -unit(PRI));
		switch (SecondaryAxisAssociation)
		{
		case PLUS_X:
		case MINUS_X:
			X = (SecondaryAxisAssociation==PLUS_X ? unit(crossp(Z, SEC)) : -unit(crossp(Z, SEC)));
			Y = crossp(Z, X);
			break;
		case PLUS_Y:
		case MINUS_Y:
			Y = (SecondaryAxisAssociation==PLUS_Y ? unit(crossp(Z, SEC)) : -unit(crossp(Z, SEC)));
			X = crossp(Y, Z);
			break;
		default:
			oapiWriteLog("unknown or invalid secondary axis association");
			throw "unknown or invalid secondary axis association";
			break;
		};
		break;
	default:
		oapiWriteLog("unknown primary axis association");
		throw "unknown primary axis association";
		break;
	};
	MATRIX3 _out = _M(X.x, X.y, X.z, Y.x, Y.y, Y.z, Z.x, Z.y, Z.z); //f* geometry! 
	return _out;
}

// orbit plane CS: 
//			origin: center of ref body
//			ref plane: plane of the orbit
//			principal direction: radius vector from the origin to the satellite
// x_op: principal direction
// y_op: 90 degrees ahead of the satellite in the sense of motion
// z_op: direction along the vector crossp(r, rdot)

// keplerian system:
//			origin: center of ref body
//			ref plane: plane of the orbit
//			principal direction: center of ref body to perifocus of satellite orbit
// x_p: principal direction
// y_p: 90 degrees ahead of the satellite in the sense of motion
// z_p: direction along the vector crossp(r, rdot)

// equinoctial system:
//			origin: center of ref body
//			ref plane: plane of the orbit
//			principal direction: center of ref body to origin of longitudes, lies in the plane of the orbit, 
//				displaced by \Omega from the ascending node N where \Omega is the RAAN.
// x_ep = f: principal direction
// y_ep = g: 90 degrees ahead of the satellite in the sense of motion
// z_ep = w: direction along the vector crossp(r, rdot)

// it is necessary to define for an arbitrary time the rotation axis of a body. 
// hence, orbiter's functions must be supplemented with others

void TwoVectorFrame(const VECTOR3& PRI, 
							  const AXISASSOCIATION PrimaryAxisAssociation, 
							  const VECTOR3& SEC,
							  const AXISASSOCIATION SecondaryAxisAssociation, 
							  VECTOR3& VX, VECTOR3& VY, VECTOR3& VZ)
{
 MATRIX3 frame = TwoVectorFrame(PRI, 
							  PrimaryAxisAssociation, 
							  SEC,
							  SecondaryAxisAssociation);
 VX = _V(frame.m11, frame.m12, frame.m13);//???
 VY = _V(frame.m21, frame.m22, frame.m23);//???
 VZ = _V(frame.m31, frame.m32, frame.m33);//???
}

inline double angsep(const VECTOR3 a, const VECTOR3 b)
{
	VECTOR3 uA = unit(a);
	VECTOR3 uB = unit(b);
	double largeA = dotp(uA, uB);
	if (largeA<0.01)
	{
		double smallA = length(crossp(uA, uB));
		return asin(smallA);
	}
	else return acos(largeA);

}

// from SSU:
static inline VECTOR3 RotateVectorX(const VECTOR3 &v, double angle) //rotates about angle (in degrees) in X-axis
{
	VECTOR3 Output;
	Output.x=v.x;
	Output.z=v.z*cos(angle*RAD)-v.y*sin(angle*RAD);
	Output.y=v.y*cos(angle*RAD)+v.z*sin(angle*RAD);
	return Output;
}
static inline VECTOR3 RotateVectorY(const VECTOR3 &v, double angle) //rotates about angle (in degrees) in Y-axis
{
	VECTOR3 Output;
	Output.y=v.y;
	Output.x=v.x*cos(angle*RAD)-v.z*sin(angle*RAD);
	Output.z=v.z*cos(angle*RAD)+v.x*sin(angle*RAD);
	return Output;
}
static inline VECTOR3 RotateVectorZ(const VECTOR3 &v, double angle) //rotates about angle (in degrees) in Z-axis
{
	VECTOR3 Output;
	Output.x=v.x*cos(angle*RAD)-v.y*sin(angle*RAD);
	Output.y=v.x*sin(angle*RAD)+v.y*cos(angle*RAD);
	Output.z=v.z;
	return Output;
}

VECTOR3 inline GetPYR2(const VECTOR3 Pitch, const VECTOR3 YawRoll)
{	
	VECTOR3 Res = { 0, 0, 0 };
	VECTOR3 P, YR;
	// Normalize the vectors
	P = unit(Pitch);
	YR = unit(YawRoll);
	VECTOR3 H = unit(crossp(P, YR));

	Res.data[YAW] = -asin(P.x);

	Res.data[ROLL] = atan2(H.x, YR.x);

	Res.data[PITCH] = atan2(P.y, P.z);
	return Res;
}

VECTOR3 inline CalcPitchYawRollAngles(const VESSEL* Spacecraft,
									  const VECTOR3 RelAttitude, 
									  const VECTOR3 RefAttitude)
{
	RefPoints GlobalPts, LocalPts;
	VECTOR3 PitchUnit = {0, 0, 1.0}, YawRollUnit = {1.0, 0, 0};

	RotateVector(PitchUnit, RelAttitude, PitchUnit);
	RotateVector(YawRollUnit, RelAttitude, YawRollUnit);

	RotateVector(PitchUnit, RefAttitude, GlobalPts.Pitch);
	RotateVector(YawRollUnit, RefAttitude, GlobalPts.Yaw);

	VECTOR3 GSpacecraftPos;
	Spacecraft->GetGlobalPos(GSpacecraftPos);

	GlobalPts.Pitch = GSpacecraftPos + GlobalPts.Pitch;
	GlobalPts.Yaw = GSpacecraftPos + GlobalPts.Yaw;	

	Spacecraft->Global2Local(GlobalPts.Pitch, LocalPts.Pitch);
	Spacecraft->Global2Local(GlobalPts.Yaw, LocalPts.Yaw);

	return GetPYR(LocalPts.Pitch, LocalPts.Yaw);
}

void inline CalcTargetRelative(const VESSEL* Spacecraft, const OBJHANDLE TargetRef, 
							   VECTOR3& RefAttitude,
							   VECTOR3& AttitudeOffset,
							   VECTOR3& RelPos, 
							   VECTOR3& RelVel, 
							   double& RadialVel,
							   VECTOR3& PitchYawRoll,
							   VECTOR3& DelPitchYawRoll)
{
	VECTOR3 SpacecraftPos, GSpacecraftPos, GSpacecraftVel, TargetPos, GTargetPos, GVel;
	VECTOR3 GRelPos, H;
	
	oapiGetGlobalPos(Spacecraft->GetHandle(), &GSpacecraftPos);
	oapiGetGlobalVel(Spacecraft->GetHandle(), &GSpacecraftVel);
	
	oapiGetGlobalPos(TargetRef, &GTargetPos); // Get position of target
	Spacecraft->Global2Local(GSpacecraftPos, SpacecraftPos); // Convert to positions to local coordinates
	Spacecraft->Global2Local(GTargetPos, TargetPos);
	RelPos = TargetPos - SpacecraftPos; // Calculate relative position of target in local coordinates
	Spacecraft->GetRelativePos(TargetRef, GRelPos); // Get position of spacecraft relative to target
	
	VESSELSTATUS Status;
	Spacecraft->GetStatus(Status);

	
	GRelPos = -GRelPos; // Reverse vector so that the it points from spacecraft to target

	if ( (length(crossp(unit(-Status.rvel), unit(GRelPos)))) < 0.1 ) { // Check to see if target is too close to velocity vector. Corresponds to approx 5deg.
		H = crossp(Status.rpos, GRelPos); // Use local vectical as roll reference
	}
	else {
		H = crossp(Status.rvel, GRelPos); // Use velocity vector as roll reference
	}

	RefAttitude = GetPYR2(GRelPos, H);
	
	DelPitchYawRoll = PitchYawRoll; //Save current pitch, yaw & roll so we can calc the change

	PitchYawRoll = CalcPitchYawRollAngles(Spacecraft, AttitudeOffset, RefAttitude);

	DelPitchYawRoll = PitchYawRoll - DelPitchYawRoll; //Calculate the change in PYR

	// Calculate relative velocity
	Spacecraft->GetRelativeVel(TargetRef, GVel);
	Spacecraft->Global2Local((GVel + GSpacecraftPos), RelVel);

	// Compute the radial component
	RadialVel = dotp(RelPos,RelVel) / length(RelPos);
}


VECTOR3 inline CalcTargetPYR(const VESSEL* Spacecraft, 
							 const OBJHANDLE TargetRef, 
							 VECTOR3& RefEulers)
{
	VECTOR3 SpacecraftPos, GSpacecraftPos, GSpacecraftVel, TargetPos, GTargetPos;//, GVel;
	VECTOR3 GRelPos, H;
	VECTOR3 AttitudeOffset = NULL_VECTOR;
	VECTOR3 PitchYawRoll;

	oapiGetGlobalPos(Spacecraft->GetHandle(), &GSpacecraftPos);
	oapiGetGlobalVel(Spacecraft->GetHandle(), &GSpacecraftVel);
	
	oapiGetGlobalPos(TargetRef, &GTargetPos); // Get position of target
	Spacecraft->Global2Local(GSpacecraftPos, SpacecraftPos); // Convert to positions to local coordinates
	Spacecraft->Global2Local(GTargetPos, TargetPos);
	//RelPos = TargetPos - SpacecraftPos; // Calculate relative position of target in local coordinates
	Spacecraft->GetRelativePos(TargetRef, GRelPos); // Get position of spacecraft relative to target
	
	VESSELSTATUS Status;
	Spacecraft->GetStatus(Status);

	
	GRelPos = -GRelPos; // Reverse vector so that the it points from spacecraft to target

	if ( (length(crossp(unit(-Status.rvel), unit(GRelPos)))) < 0.1 ) { // Check to see if target is too close to velocity vector. Corresponds to approx 5deg.
		H = crossp(Status.rpos, GRelPos); // Use local vectical as roll reference
	}
	else {
		H = crossp(Status.rvel, GRelPos); // Use velocity vector as roll reference
	}

	RefEulers = GetPYR2(GRelPos, H);
	
	//DelPitchYawRoll = PitchYawRoll; //Save current pitch, yaw & roll so we can calc the change

	PitchYawRoll = CalcPitchYawRollAngles(Spacecraft, AttitudeOffset, RefEulers);

	return PitchYawRoll;

	//DelPitchYawRoll = PitchYawRoll - DelPitchYawRoll; //Calculate the change in PYR

	// Calculate relative velocity
	//Spacecraft->GetRelativeVel(TargetRef, GVel);
	//Spacecraft->Global2Local((GVel + GSpacecraftPos), RelVel);

	//// Compute the radial component
	//RadialVel = dotp(RelPos,RelVel) / length(RelPos);
}

inline MATRIX3 _RotMatrix(const ROT_AXIS axis, const double angle_Rad)
{
	double ct = cos(angle_Rad);
	double st = sin(angle_Rad);

	double ux = 0.0;
	double uy = 0.0;
	double uz = 0.0;
	double ux2 = 0.0;
	double uy2 = 0.0;
	double uz2 = 0.0;

	switch (axis)
	{
	case AXIS_X:
		ux = 1.0;
		ux2 = 1.0;
		break;
	case AXIS_Y:
		uy = 1.0;
		uy2 = 1.0;
		break;
	case AXIS_Z:
		uz = 1.0;
		uz2 = 1.0;
		break;
	default:
		oapiWriteLog("No such axis");
		throw "No such axis";
		break;
	};

	MATRIX3 RG = _M(ux2+(1-ux2)*ct,      ux*uy*(1-ct)-uz*st, ux*uz*(1-ct)+uy*st,
		ux*uy*(1-ct)+uz*st,  uy2+(1-uy2)*ct,     uy*uz*(1-ct)-ux*st,
		ux*uz*(1-ct)-uy*st,  uy*uz*(1-ct)+ux*st, uz2+(1-uz2)*ct);
	return RG;
}

inline MATRIX3 _RotMatrix(const ROT_AXIS axis1, 
						  const ROT_AXIS axis2, 
						  const double angle1_rad, 
						  const double angle2_rad)
{
	MATRIX3 RG = mul(_RotMatrix(axis2, angle2_rad), _RotMatrix(axis1, angle1_rad));
	return RG;
}

inline MATRIX3 _RotMatrix(const ROT_AXIS axis1, 
						  const ROT_AXIS axis2, 
						  const ROT_AXIS axis3, 
						  const double angle1_rad, 
						  const double angle2_rad,
						  const double angle3_rad)
{
	MATRIX3 RG = mul(
		_RotMatrix(axis3, angle3_rad),
		mul(
		_RotMatrix(axis2, angle2_rad),
		_RotMatrix(axis1, angle1_rad)
		)
		);
	return RG;
}

inline MATRIX3 transpose(const MATRIX3 M)
{
	return _M(M.m11, M.m21, M.m31, 
			  M.m12, M.m22, M.m32,
			  M.m13, M.m23, M.m33);
}

inline MATRIX3 operator+(const MATRIX3& m1, const MATRIX3& m2)
{
	MATRIX3 result;
	for (int i = 0; i<9; i++)
	{
		result.data[i] = m1.data[i]+m2.data[i];
	};
	return result;
}

inline MATRIX3 operator-(const MATRIX3& m1, const MATRIX3& m2)
{
	MATRIX3 result;
	for (int i = 0; i<9; i++)
	{
		result.data[i] = m1.data[i]-m2.data[i];
	};
	return result;
}

inline double trace(const MATRIX3 M)
{
	return M.m11+M.m22+M.m33;
}



// all in radians!
inline void FindAngleAndAxis_KALCMANU(const MATRIX3 InitialAttitude, 
									  const MATRIX3 DesiredAttitude,
									  VECTOR3& RotationAxis, 
									  double& RotAngle)
{
	MATRIX3 R_CB = mul(transpose(InitialAttitude), DesiredAttitude);
	MATRIX3 R_CB_T = transpose(R_CB);
	MATRIX3 R_A = 0.5*(R_CB-R_CB_T);
	double cosRotAngle = 0.5*(trace(R_CB)-1.0);
	if (fabs(cosRotAngle)>1.0)
	{
		oapiWriteLog("arccos argument modulus greater than unity");
		throw "arccos argument modulus greater than unity"; // shouldn't be here!
	};
	RotAngle = acos(cosRotAngle);
	double negRA6 = -R_A.m23;
	double RA3 = R_A.m13;
	double negRA2 = -R_A.m12;
	if (RotAngle*DEG>170.0)
	{
		MATRIX3 R_S = 0.5*(R_CB+R_CB_T);
		double RS6 = R_S.m23;
		double RS3 = R_S.m13;
		double RS2 = R_S.m12;
		// check!
		VECTOR3 _RotAxis = _V(sqrt((R_S.m11-cosRotAngle)/(1.0-cosRotAngle)),
							  sqrt((R_S.m22-cosRotAngle)/(1.0-cosRotAngle)),
							  sqrt((R_S.m33-cosRotAngle)/(1.0-cosRotAngle)));
		int largest = -1;
		if (_RotAxis.x<_RotAxis.y)
		{
			largest = (_RotAxis.x<_RotAxis.z ? 2:0);
		}
		else
		{
			largest = (_RotAxis.y<_RotAxis.z ? 2:1);
		};
		switch (largest)
		{
		case 0: //x
			if (negRA6<0.0) _RotAxis = -_RotAxis;
			if (RS2<0.0) _RotAxis.y = -_RotAxis.y;
			if (RS3<0.0) _RotAxis.z = -_RotAxis.z;
			break;
		case 1: //y
			if (RA3<0.0) _RotAxis = -_RotAxis;
			if (RS2<0.0) _RotAxis.x = -_RotAxis.x;
			if (RS6<0.0) _RotAxis.z = -_RotAxis.z;
			break;
		case 2: //z
			if (negRA2<0.0) _RotAxis = -_RotAxis;
			if (RS3<0.0) _RotAxis.x = -_RotAxis.x;
			if (RS6<0.0) _RotAxis.y = -_RotAxis.y;
			break;
		default:
			oapiWriteLog("fall-through");
			throw "fall-through";
			break;
		};
		for (int i = 0; i<3; i++) RotationAxis.data[i] = _RotAxis.data[i];
	}
	else
	{
		RotationAxis = _V(negRA6, RA3, negRA2); // can convert to quaternion?
	};
}
// angular velocity vector should be parallel to the rotation axis! 
//the solution: ang velocities components should match the axis while far away from the terminal point.

inline void FindAngleAndAxis_KALCMANU(const VECTOR3 InitialAttitude, 
									  const VECTOR3 DesiredAttitude,
									  VECTOR3& RotationAxis, 
									  double& RotAngle)
{
	MATRIX3 IA = EulerToMatrix(InitialAttitude);
	MATRIX3 DA = EulerToMatrix(DesiredAttitude);
	FindAngleAndAxis_KALCMANU(IA, DA, RotationAxis, RotAngle);
}

inline double CylinderProjectionArea(const double alpha, const double L, const double diam)
{
	// to calculate heat load
	// TBD
	// cos alpha! 
	// 
	double S_rectangle = L*diam;
	double S_circle = PI*diam*diam*0.25;
	double cosA = cos(alpha);
	if (cosA<0.0) cosA = -cosA;
	double cosA_n = cos(alpha+HALFPI);
	if (cosA_n<0.0) cosA_n = -cosA_n;//?
	return S_circle*cosA+S_rectangle*cosA_n;
}
// need: nadir angle 

///RA, Dec in Degrees
inline VECTOR3 BrianJ_RADecToEulers(const double RA, const double Dec)
{

	double ecliptic_inc = J2000_Ecliptic_Obliquity*DEG_TO_RAD;

	double dec_angle = Dec*DEG_TO_RAD;
	double ra_angle = RA*DEG_TO_RAD;

	////// vessel axes in equatorial frame
	VECTOR3 vessel_x = _V(1,0,0);
	VECTOR3 vessel_y = _V(0,1,0);
	VECTOR3 vessel_z = _V(0,0,1);

	////// create rotation matrix for rotation by declination angle
	MATRIX3 RG_result;

	RG_result = _RotMatrix(AXIS_X, AXIS_Y, AXIS_X, -dec_angle, (PI/2)-ra_angle, ecliptic_inc);

	vessel_x = mul(RG_result,vessel_x);
	vessel_y = mul(RG_result,vessel_y);
	vessel_z = mul(RG_result,vessel_z);

	double gamma = atan2(vessel_x.x, vessel_y.x);
	double beta = -1.0*(asin(vessel_z.x));
	double alpha = atan2(vessel_z.y, vessel_z.z);

	return _V(alpha, beta, 0.0);
}

// from SSU
inline double NullStartAngle(double Rate, double Mass, double Moment, double Torque)
{
	//???
	if(fabs(Rate)>0.0001) 
	{
		double Time = (Mass*Moment*Rate)/Torque;
		double Angle=0.5*Rate*Time;
		return DEG*Angle;
	}
	else return 0.0;
}

//inline bool IsSpacecraftWithinLOS(OBJHANDLE obj, double Lat, double Lon)
//{
//   local retv
//   double rs_angle;// -- rbody to sun angle
//   double rh_angle; //-- rbody to horizon angle
//   VECTOR3 rppos;
//   VECTOR3 rspos;
//   VESSEL* obj_v = oapiGetVesselInterface(obj);
//   if (obj_v == NULL) return false;
//   OBJHANDLE body = obj_v->GetSurfaceRef();
//   obj_v->GetRelativePos(body, rppos);
//   double rplen = length(rppos);
//   rh_angle = asin(oapiGetSize(body)/rplen)
//   normalise(rppos);
//   local star = oapi.get_objhandle('Sun')
//   rspos = obj_v:get_relativepos(star) -- sun position
//   rspos = vec.div(rspos,vec.length(rspos)) -- normalise
//   rs_angle = math.acos(vec.dotp(rppos,rspos))+(50/60*RAD)
//   if (rs_angle > rh_angle) then
//     retv = 1
//   else
//     retv = 0
//   end
//   return retv
//}

#endif //! __ROTATIONS_H_